{$I cHeader.inc}
unit cHTTP;

interface

uses
  // System units
  SysUtils,
  Classes,
  scktcomp,

  // L0 units
  cInternetStandards,
  cStream;



{                                                                              }
{ HTTP Protocol v0.04                                                          }
{                                                                              }
{   Implementation of the HTTP Protocol for clients and servers.               }
{                                                                              }
{ This unit is copyrighted  2000 by David Butler (david@e.co.za)              }
{                                                                              }
{ I invite you to use this unit, free of charge.                               }
{ I invite you to distibute this unit, but it must be for free.                }
{ I also invite you to contribute to its development, but do not distribute    }
{ a modified copy of this file, send modifications, suggestions and bug        }
{ reports to david@e.co.za                                                     }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   2000/03/11  v0.01  Initial version.                                        }
{   2000/03/26  v0.02  Exception handling for ThttpClientServerThread.         }
{   2000/04/28  v0.03  Moved HTTP code from cInternetStandards.                }
{   2000/05/07  v0.04  Added ThttpMessage.                                     }
{                                                                              }
{ From RFC2616 (Hypertext Transfer Protocol -- HTTP/1.1)                       }
{ From RFC2396 (Uniform Resource Identifiers (URI): Generic Syntax)            }
{                                                                              }


const
  { HTTP Protocol version numbers }
  http_MajorVersion = 1;
  http_MinorVersion = 1;

const
  { HTTP Status code classes }
  http_Informational        = 1;     // 1xx Informational
  http_Successful           = 2;     // 2xx Successful
  http_Redirection          = 3;     // 3xx Redirection
  http_ClientError          = 4;     // 4xx Client Error
  http_ServerError          = 5;     // 5xx Server Error

  { HTTP Status codes }
  http_Continue             = 100;
  http_OK                   = 200;
  http_Created              = 201;
  http_Accepted             = 202;
  http_NoContent            = 204;
  http_ResetContent         = 205;
  http_PartialContent       = 206;
  http_Found                = 302;
  http_SeeOther             = 303;
  http_NotModified          = 304;
  http_UseProxy             = 305;
  http_TemporaryRedirect    = 307;
  http_BadRequest           = 400;
  http_Unauthorized         = 401;
  http_Forbidden            = 403;
  http_NotFound             = 404;
  http_MethodNotFound       = 405;
  http_RequestTimeout       = 408;
  http_Gone                 = 410;
  http_LengthRequired       = 411;
  http_EntityTooLarge       = 413;
  http_URITooLong           = 414;
  http_UnsupportedMediaType = 415;
  http_InternalServerError  = 500;
  http_NotImplemented       = 501;
  http_VersionNotSupported  = 505;


{ Ehttp                                                                        }
type
  Ehttp = class (Exception)
    httpStatusCode : Integer;
    Constructor Create (const StatusCode : Integer; const Msg : String);
  end;


{                                                                              }
{ ThttpMessage                                                                 }
{   Base class for client and server messages.                                 }
{                                                                              }
type
  ThttpMessage = class
    protected
    FMajorVersion : Integer;
    FMinorVersion : Integer;
    FHeader       : THeader;
    FIncludeDate  : Boolean;

    Procedure PrepareHeader; virtual;

    public
    Constructor Create (const MajorVersion : Integer = http_MajorVersion;
                        const MinorVersion : Integer = http_MinorVersion);
    Destructor Destroy; override;

    Property MajorVersion : Integer read FMajorVersion write FMajorVersion;
    Property MinorVersion : Integer read FMinorVersion write FMinorVersion;
    Property IncludeDate : Boolean read FIncludeDate write FIncludeDate;
    Property Header : THeader read FHeader write FHeader;

    Function GetAsString : String; virtual; abstract;
  end;



{                                                                              }
{ ThttpRequest                                                                 }
{                                                                              }
type
  ThttpRequest = class (ThttpMessage)
    protected
    FMethod : String;
    FURI    : String;

    public
    Constructor Create (const Method, URI : String;
                const MajorVersion : Integer = http_MajorVersion;
                const MinorVersion : Integer = http_MinorVersion);

    Function GetAsString : String; override;
  end;



{                                                                              }
{ ThttpResponse                                                                }
{                                                                              }
type
  ThttpResponse = class (ThttpMessage)
    protected
    FStatusCode     : Integer;
    FContent        : String;
    FIncludeContent : Boolean;
    FContentType    : String;
    FServerName     : String;

    Procedure SetContent (const Content : String);
    Procedure SetStatusCode (const StatusCode : Integer);
    Procedure PrepareHeader; override;

    public
    Constructor Create (const StatusCode : Integer = http_OK;
                const MajorVersion : Integer = http_MajorVersion;
                const MinorVersion : Integer = http_MinorVersion);

    Property StatusCode : Integer read FStatusCode write SetStatusCode;
    Property IncludeContent : Boolean read FIncludeContent write FIncludeContent;
    Property Content : String read FContent write SetContent;
    Property ContentType : String read FContentType write FContentType;
    Property ServerName : String read FServerName write FServerName;

    Function GetAsString : String; override;
  end;


{                                                                              }
{ ThttpServer                                                                  }
{                                                                              }
type
  TLogType = (lt_Socket, lt_httpServer);
  TOnhttpServerLog = Procedure (const LogType : TLogType; const LogMsg : String) of object;
  ThttpServerClientThread = class;
  TOnhttpServerRequest = Function (const Sender : ThttpServerClientThread;
    const Method, URI : String; const Hdr : THeader) : ThttpResponse of object;
  EhttpServer = class (Exception);
  ThttpServer = class (TComponent)
    protected
    FActive           : Boolean;
    FListenPort       : Integer;
    FServerSocket     : TServerSocket;
    FOnLog            : TOnhttpServerLog;
    FOnRequest        : TOnhttpServerRequest;
    FTimeOut          : Integer;

    Procedure OnGetThread (Sender : TObject; ClientSocket : TServerClientWinSocket;
              var SocketThread : TServerClientThread);
    Procedure SetActive (const Active : Boolean);
    Procedure SetListenPort (const ListenPort : Integer);
    Procedure Log (const LogType : TLogType; const LogMsg : String);
    Procedure RaiseError (const Msg : String);

    public
    Constructor Create (AOwner : TComponent); override;
    Destructor Destroy; override;

    Property Active : Boolean read FActive write SetActive;
    Property ListenPort : Integer read FListenPort write SetListenPort;
    Property OnLog : TOnhttpServerLog read FOnLog write FOnLog;
    Property OnRequest : TOnhttpServerRequest read FOnRequest write FOnRequest;
    Property TimeOut : Integer read FTimeOut write FTimeOut;
  end;

  { ThttpServerClientThread                                                    }
  ThttpServerClientThread = class (TServerClientThread)
    protected
    FServer  : ThttpServer;
    FClient  : TStream;
    FOnLog   : TOnhttpServerLog;
    FTimeOut : Integer;
    FLogType : TLogType;
    FLogMsg  : String;

    Procedure SynchronisedLog;
    Procedure Log (const LogType : TLogType; const LogMsg : String);

    public
    Constructor Create (const Server : ThttpServer;
                const ClientSocket : TServerClientWinSocket);
    Procedure ClientExecute; override;
    Procedure DoTerminate; override;

    Property OnLog : TOnhttpServerLog read FOnLog write FOnLog;
    Property TimeOut : Integer read FTimeOut write FTimeOut;
  end;




Function  httpDecodeURI (const URI : String) : String;
Procedure httpDecodeRequestLine (const R : String;
                                var MajorVersion, MinorVersion : Integer;
                                var Method, URI : String);
Procedure httpDecodeURL (const URL : String; var Protocol, Address, Path : String);
Function  ContentTypeFromExtention (Extention : String) : String;


implementation



uses
  // L0 units
  cUtil,
  cMaths,
  cStrings,
  cDateTime;     // High-precision timer

const
  CRLF = ASCII_CR + ASCII_LF;
  SP   = ASCII_SP;

{ HTTP Status Codes Encoding                                                   }
Function httpStatus (const StatusCode : Integer; const Comment : String = '') : String;
type
  ThttpStatusCode = record
    Code   : Integer;
    Status : String;
  end;
const
  httpStatusCodes = 41;
  httpStatusCode : Array [1..httpStatusCodes] of ThttpStatusCode = (
      (Code: http_Continue;              Status: 'Continue'),
      (Code: 101;                        Status: 'Switching Protocols'),
      (Code: http_OK;                    Status: 'OK'),
      (Code: http_Created;               Status: 'Created'),
      (Code: http_Accepted;              Status: 'Accepted'),
      (Code: 203;                        Status: 'Non-Authoritative Information'),
      (Code: http_NoContent;             Status: 'No Content'),
      (Code: http_ResetContent;          Status: 'Reset Content'),
      (Code: http_PartialContent;        Status: 'Partial Content'),
      (Code: 300;                        Status: 'Multiple Choices'),
      (Code: 301;                        Status: 'Moved Permanently'),
      (Code: http_Found;                 Status: 'Found'),
      (Code: http_SeeOther;              Status: 'See Other'),
      (Code: http_NotModified;           Status: 'Not modified'),
      (Code: http_UseProxy;              Status: 'Use Proxy'),
      (Code: 306;                        Status: 'Reserved'),
      (Code: http_TemporaryRedirect;     Status: 'Temporary Redirect'),
      (Code: http_BadRequest;            Status: 'Bad Request'),
      (Code: http_Unauthorized;          Status: 'Unauthorized'),
      (Code: 402;                        Status: 'Payment Required'),
      (Code: http_Forbidden;             Status: 'Forbidden'),
      (Code: http_NotFound;              Status: 'Not Found'),
      (Code: http_MethodNotFound;        Status: 'Method Not Found'),
      (Code: 406;                        Status: 'Not Acceptable'),
      (Code: 407;                        Status: 'Proxy Authentication Required'),
      (Code: http_RequestTimeout;        Status: 'Request Timeout'),
      (Code: 409;                        Status: 'Conflict'),
      (Code: http_Gone;                  Status: 'Gone'),
      (Code: http_LengthRequired;        Status: 'Length Required'),
      (Code: 412;                        Status: 'Precondition Failed'),
      (Code: http_EntityTooLarge;        Status: 'Request Entity Too Large'),
      (Code: http_URITooLong;            Status: 'Request-URI Too Long'),
      (Code: http_UnsupportedMediaType;  Status: 'Unsupported Media Type'),
      (Code: 416;                        Status: 'Requested Range Not Satisfiable'),
      (Code: 417;                        Status: 'Expectation Failed'),
      (Code: http_InternalServerError;   Status: 'Internal Server Error'),
      (Code: http_NotImplemented;        Status: 'Not Implemented'),
      (Code: 502;                        Status: 'Bad Gateway'),
      (Code: 503;                        Status: 'Service Unavailable'),
      (Code: 504;                        Status: 'Gateway Timeout'),
      (Code: http_VersionNotSupported;   Status: 'HTTP Version Not Supported'));
var I : Integer;
  Begin
    Result := IntToStr (StatusCode);
    For I := 1 to httpStatusCodes do
      if httpStatusCode [I].Code = StatusCode then
        begin
          Result := Result + ' ' + httpStatusCode [I].Status;
          break;
        end;
    if Comment <> '' then
      Result := Result + ' (' + Comment + ')';
  End;

Constructor Ehttp.Create (const StatusCode : Integer; const Msg : String);
  Begin
    inherited Create (httpStatus (StatusCode, Msg));
    httpStatusCode := StatusCode;
  End;


{ HTTP Version Number Encoding                                                 }
Function httpVersionNumber (const MajorVersion : Integer = http_MajorVersion;
                            const MinorVersion : Integer = http_MinorVersion) : String;
  Begin
    Result := 'HTTP/' + IntToStr (MajorVersion) + '.' + IntToStr (MinorVersion);
  End;


{ ThttpMessage                                                                 }
Constructor ThttpMessage.Create (const MajorVersion : Integer = http_MajorVersion; const MinorVersion : Integer = http_MinorVersion);
  Begin
    inherited Create;
    FMajorVersion := MajorVersion;
    FMinorVersion := MinorVersion;
    FIncludeDate := True;
    FHeader := THeader.Create;
  End;

Destructor ThttpMessage.Destroy;
  Begin
    FreeAndNil (FHeader);
    inherited Destroy;
  End;

Procedure ThttpMessage.PrepareHeader;
  Begin
    FHeader ['Date'] := AHeaderField (Cond (FIncludeDate, TDateField.Create (Now), nil));
  End;


{ ThttpRequest                                                                 }
Constructor ThttpRequest.Create (const Method, URI : String; const MajorVersion : Integer = http_MajorVersion; const MinorVersion : Integer = http_MinorVersion);
  Begin
    inherited Create (MajorVersion, MinorVersion);
    FMethod := Method;
    FURI := URI;
  End;

Function ThttpRequest.GetAsString : String;
  Begin
    PrepareHeader;
    Result := FMethod + SP + FURI + SP + httpVersionNumber (FMajorVersion, FMinorVersion) + CRLF +
              FHeader.AsString + CRLF;
  End;


{ From RFC2396 (Uniform Resource Identifiers (URI): Generic Syntax)            }
Function httpDecodeURI (const URI : String) : String;
var I, J : Integer;
    R    : Boolean;
  Begin
    J := 0;
    Repeat
      I := PosNext ('%', URI, J);
      R := (I > 0) and (I < Length (URI) - 1);
      if R then
        begin
          Result := Result + CopyRange (URI, J + 1, I - 1);
          try
            Result := Result + Char (HexToInt (Copy (URI, I + 1, 2)));
          except
            Result := Result + Copy (URI, I, 3);
          end;
          J := I + 2;
        end;
    Until not R;
    Result := Result + CopyFrom (URI, J + 1);
  End;

Function httpDecodeFileLocation (const Path : String) : String;
  Begin
    Result := Replace ('/', '\', TrimLeft (Path, ['/']));
    if (Length (Result) >= 2) and (Result [2] = '|') then
      Result [2] := ':';
  End;

Procedure httpDecodeURL (const URL : String; var Protocol, Address, Path : String);
var S : String;
    I : Integer;
  Begin
    Address := '';
    Path := '';
    S := httpDecodeURI (URL);
    I := PosNext (':', S);
    if I = 0 then // no protocol
      Protocol := '' else
      begin
        Protocol := LowerCase (CopyLeft (S, I - 1));
        S := CopyFrom (S, I + 1);
      end;

    if Protocol = 'file' then
      Path := httpDecodeFILELocation (S) else
    if Protocol = 'mail' then
      Address := S else
    if (Protocol = 'ftp') or (Protocol = 'http') then
      if Match ('//', S) then
        begin
          Address := CopyBefore (S, '/', True, 3);
          Path := CopyFrom (S, '/', False, 3);
        end else
        Path := S
    else
      Path := S;
  End;

{ From RFC2616 (Hypertext Transfer Protocol -- HTTP/1.1) 5.1                   }
{   Methods: GET, OPTIONS, HEAD, POST, PUT, DELETE, TRACE, CONNECT             }
{   Returns MajorVersion, MinorVersion = 0 if request not in a valid format.   }
Procedure httpDecodeRequestLine (const R : String;
         var MajorVersion, MinorVersion : Integer;
         var Method, URI : String);
var S, T : String;
    I    : Integer;
  Begin
    // Check version number
    S := Trim (R, SPACE);
    MajorVersion := 0;
    MinorVersion := 0;
    I := PosPrev (SPACE, S);
    if I > 0 then
      begin
        T := UpperCase (CopyFrom (S, I + 1));
        if Match ('HTTP/', T, 1) then
          begin
            S := Trim (CopyLeft (S, I - 1), SPACE);
            T := CopyFrom (T, 6);
            I := PosNext ('.', T);
            if I > 0 then
              try
                MajorVersion := StrToInt (CopyLeft (T, I - 1));
                MinorVersion := StrToInt (CopyFrom (T, I + 1));
              except end;
          end;
      end;

    // Method & URI
    Method := UpperCase (CopyBefore (S, SPACE, True));
    URI := httpDecodeURI (Trim (CopyAfter (S, SPACE), SPACE));
  End;

{ Status-Line = HTTP-Version SP Status-Code SP Reason-Phrase CRLF              }
Function httpStatusLine (const StatusCode : Integer; const MajorVersion : Integer = http_MajorVersion;
                            const MinorVersion : Integer = http_MinorVersion) : String;
  Begin
    Result := httpVersionNumber (MajorVersion, MinorVersion) +
              ' ' +
              httpStatus (StatusCode) +
              CRLF;
  End;

Function ContentTypeFromExtention (Extention : String) : String;
const
  // ContentType
  ct_HTML        = 'text/html; charset=ISO-8859-4';
  ct_ASCIIText   = 'text/plain; charset=US-ASCII';
  ct_JPG         = 'image/jpeg';
  ct_GIF         = 'image/gif';
  ct_BMP         = 'image/x-xbitmap';
  ct_PNG         = 'image/png';
  ct_Binary      = 'application/binary';
type
  TContentTypeEntry = record
    Ext : String;
    CT  : String;
  end;
const
  BuiltInContentTypes = 8;
  BuiltInContentType : Array [1..BuiltInContentTypes] of TContentTypeEntry = (
      (Ext: '.htm'; CT: ct_HTML),     (Ext: '.html'; CT: ct_HTML),
      (Ext: '.jpg'; CT: ct_JPG),      (Ext: '.jpeg'; CT: ct_JPG),
      (Ext: '.gif'; CT: ct_JPG),      (Ext: '.png';  CT: ct_PNG),
      (Ext: '.bmp'; CT: ct_BMP),      (Ext: '.txt';  CT: ct_ASCIIText) );
var I : Integer;
  Begin
    Extention := LowerCase (Extention);
    For I := 1 to BuiltInContentTypes do
      if Extention = BuiltInContentType [I].Ext then
        begin
          Result := BuiltInContentType [I].CT;
          exit;
        end;
    Result := '';
  End;

{ ThttpResponse                                                                }
Constructor ThttpResponse.Create (const StatusCode : Integer; const MajorVersion : Integer; const MinorVersion : Integer);
  Begin
    inherited Create (MajorVersion, MinorVersion);
    SetStatusCode (StatusCode);
    ServerName := 'Delphi-Fundamentals-L0-http/0.1';
  End;

Procedure ThttpResponse.PrepareHeader;
  Begin
    inherited PrepareHeader;
    FHeader ['Server'] := AHeaderField (Cond (FServerName <> '',
            THeaderField.Create ('Server', ServerName), nil));
    FHeader ['Content-Type'] := AHeaderField (Cond (ContentType <> '',
            THeaderField.Create ('Content-Type', ContentType), nil));
    FHeader ['Content-Length'] := AHeaderField (Cond (
            (StatusCode div 100 <> http_Informational) and
            (StatusCode <> http_NoContent) and
            (StatusCode <> http_NotModified),
            THeaderField.Create ('Content-Length', IntToStr (Length (Content))), nil));
  End;

Procedure ThttpResponse.SetContent (const Content : String);
  Begin
    FContent := Content;
    FIncludeContent := True;
  End;

Procedure ThttpResponse.SetStatusCode (const StatusCode : Integer);
  Begin
    FStatusCode := StatusCode;
    FIncludeDate := StatusCode div 100 <> http_Informational;
  End;

Function ThttpResponse.GetAsString : String;
  Begin
    PrepareHeader;

    Result := httpStatusLine (StatusCode) + FHeader.AsString;

    if IncludeContent then
      Result := Result + Content;
  End;

{ ThttpServer                                                                  }
Constructor ThttpServer.Create (AOwner : TComponent);
  Begin
    inherited Create (AOwner);
    FListenPort := 80;
    FTimeOut := 60 * 1000; // One minute
  End;

Destructor ThttpServer.Destroy;
  Begin
    Active := False;
    inherited Destroy;
  End;

Procedure ThttpServer.OnGetThread (Sender : TObject; ClientSocket: TServerClientWinSocket; var SocketThread: TServerClientThread);
var T : ThttpServerClientThread;
  Begin
    // Create new thread
    T := ThttpServerClientThread.Create (self, ClientSocket);
    T.OnLog := FOnLog;
    T.TimeOut := FTimeOut;
    SocketThread := T;

    // Thread cache management
    if FServerSocket.Socket.ActiveThreads > FServerSocket.ThreadCacheSize then
      FServerSocket.ThreadCacheSize := FServerSocket.Socket.ActiveThreads;

    Log (lt_httpServer, 'New thread created.');
  End;

Procedure ThttpServer.SetActive (const Active : Boolean);
  Begin
    if FActive = Active then
      exit;

    if Active then
      begin
        try
          FServerSocket := TServerSocket.Create (self);
          FServerSocket.ServerType := stThreadBlocking;
          FServerSocket.Port := FListenPort;
          FServerSocket.OnGetThread := OnGetThread;
          FServerSocket.Active := True;
        except
          on E : Exception do RaiseError ('Could not create server socket: ' + ExceptionToStr (E));
        end;
        FActive := True;
        Log (lt_httpServer, 'Server socket activated, listening on port ' + IntToStr (FListenPort) + '.');
      end else
      begin
        Log (lt_httpServer, 'Destroying server socket');
        try
          FServerSocket.Active := False;
        except
          on E : Exception do Log (lt_httpServer, 'Error deactivating socket (' + ExceptionToStr (E) + ')');
        end;
        FServerSocket.Free;
        FServerSocket := nil;
        FActive := False;
      end;
  End;

Procedure ThttpServer.SetListenPort (const ListenPort : Integer);
var R : Boolean;
  Begin
    if ListenPort = FListenPort then
      exit;

    try
      R := Active;
      try
        if R then
          SetActive (False);
        FListenPort := ListenPort;
      finally
        if R then
          SetActive (True);
      end;
    except
      on E : Exception do RaiseError ('Could not set server port to ' + IntToStr (ListenPort) + ': ' + ExceptionToStr (E));
    end;
    Log (lt_httpServer, 'Listen port set to ' + IntToStr (ListenPort));
  End;

Procedure ThttpServer.Log (const LogType : TLogType; const LogMsg : String);
  Begin
    if Assigned (FOnLog) then
      FOnLog (LogType, LogMsg);
  End;

Procedure ThttpServer.RaiseError (const Msg : String);
  Begin
    raise EhttpServer.Create (Msg);
  End;

{ ThttpServerClientThread                                                      }
Constructor ThttpServerClientThread.Create (const Server : ThttpServer; const ClientSocket : TServerClientWinSocket);
  Begin
    inherited Create (False, ClientSocket);
    FServer := Server;
    KeepInCache := True;
  End;

Procedure ThttpServerClientThread.ClientExecute;

  Function ReadLine (var S : String) : Boolean;
    Begin
      S := '';
      Result := False;

      if Terminated then
        begin
          Log (lt_httpServer, 'Thread terminated.');
          exit;
        end;
      try
        S := FClient.ExtractLine;
        Result := True;
      except
        on E : Exception do
          Log (lt_httpServer, 'Read error (' + ExceptionToStr (E) + ')');
      end;
    End;

  Procedure SendResponse (const Response : ThttpResponse);
  var S : String;
      I : Int64;
    Begin
      try try
        S := Response.GetAsString;
        I := StartTimer;
        FClient.Write (S);
        Log (lt_httpServer, 'Sent HTTP response: Code=' + IntToStr (Response.StatusCode) +
                            ', Size=' + IntToStr (Length (S)) + ', Transfer rate=' +
                            TransferRate (Length (S), MilliSecondsElapsed (I)) +
                            ', Response=' + CopyLeft (Response.GetAsString, 200));
      except
        on E : Exception do Log (lt_httpServer, 'Error sending HTTP response code ' + IntToStr (Response.StatusCode) + ' (' + ExceptionToStr (E) + ')');
      end;
      finally
        Response.Free;
      end;
    End;

var S, Method, URI : String;
    MajVer, MinVer : Integer;
    Response       : ThttpResponse;
    Hdr            : THeader;
    RespStartTime,
    StartTime      : Int64;

  Begin
    StartTime := StartTimer;
    try
      Log (lt_httpServer, 'New client (' + ClientSocket.RemoteAddress + ' - ' + ClientSocket.RemoteHost + ')');
      FClient := TSocketStream.Create (ClientSocket, FTimeOut);
    except
      on E : Exception do Log (lt_httpServer, 'Client gone (' + ExceptionToStr (E) + ')');
    end;
    try try
      FClient.EOLType := [eol_EOF, eol_CRLF];
      // Request service loop
      While not Terminated and ClientSocket.Connected do
        begin
          // Wait for a request
          try
            FClient.WaitOctetAvailable (FTimeOut);
          except
            on E : Exception do
              begin
                Log (lt_httpServer, 'Error waiting for request (' + ExceptionToStr (E) + ')');
                break; // request timed out
              end;
          end;

          if (FClient.OctetsAvailable = 0) or not ClientSocket.Connected then
            begin
              Log (lt_httpServer, 'No request' + Cond (not ClientSocket.Connected, ' - Not connected', ''));
              break;
            end;

          // Get request line
          Log (lt_httpServer, 'Reading request.');
          if not ReadLine (S) then
            break;
          httpDecodeRequestLine (S, MajVer, MinVer, Method, URI);
          Log (lt_httpServer, 'Decoded HTTP' + IntToStr (MajVer) + '.' + IntToStr (MinVer) +
                ': ' + Method + ' ' + URI);

          // Get request header (if HTTP/1 compliant request)
          if MajVer >= 1 then
            begin
              Hdr := THeader.Create;
              Repeat
                ReadLine (S);
                if S <> '' then
                  Hdr.AddField (S);
              Until S = '';
              Log (lt_httpServer, 'Header: ' + Hdr.AsString);
            end;

          // Get response
          RespStartTime := StartTimer;
          if Assigned (FServer.FOnRequest) then
            Response := FServer.FOnRequest (self, Method, URI, Hdr) else
            Response := nil;
          if not Assigned (Response) then
            Response := ThttpResponse.Create (http_NotFound);
          Log (lt_httpServer, 'Prepared response in ' + IntToStr (MilliSecondsElapsed (RespStartTime)) + 'ms.');

          // Respond
          SendResponse (Response);
          Log (lt_httpServer, 'Responded in ' + IntToStr (MilliSecondsElapsed (RespStartTime)) + 'ms');
        end;
      Log (lt_httpServer, 'Finished servicing client requests (' + IntToStr (MillisecondsElapsed (StartTime)) + 'ms).');
    except
      on E : Ehttp do
        begin
          Log (lt_httpServer, 'HTTP Error (' + ExceptionToStr (E) + ')');
          SendResponse (ThttpResponse.Create (E.httpStatusCode));
        end;
      on E : Exception do
        begin
          Log (lt_httpServer, 'Exception (' + ExceptionToStr (E) + ')');
          Response := ThttpResponse.Create (http_OK);
          Response.Content := htmlDocument ('Internal error', htmlBold (E.ClassName + ': ') + E.Message);
          SendResponse (Response);
        end;
    end;
    finally
      try
        ClientSocket.Close;
        Log (lt_httpServer, 'Socket closed.');
      except
        on E : Exception do Log (lt_httpServer, 'Error closing socket: ' + ExceptionToStr (E));
      end;
      FreeAndNil (FClient);
    end;
  End;

Procedure ThttpServerClientThread.DoTerminate;
  Begin
    Log (lt_httpServer, 'Terminating thread.');
    inherited DoTerminate;
  End;

{ The log event is called using the Synchronise method of the thread, ie the   }
{ log event won't be re-entered.                                               }
Procedure ThttpServerClientThread.SynchronisedLog;
  Begin
    FOnLog (FLogType, FLogMsg);
  End;

Procedure ThttpServerClientThread.Log (const LogType : TLogType; const LogMsg : String);
  Begin
    if Assigned (FOnLog) then
      begin
        FLogType := LogType;
        FLogMsg := LogMsg;
        Synchronize (SynchronisedLog);
      end;
  End;

end.






