{$I cHeader.inc}
unit cInternetStandards;

interface

uses
  // System units
  SysUtils,
  Controls,      // TControl
  Graphics,      // TColor

  // L0 units
  cStrings,
  cStream;



{                                                                              }
{ Internet Standards v0.11                                                     }
{                                                                              }
{   Implementation of various RFC standards                                    }
{                                                                              }
{ This unit is copyrighted  2000 by David Butler (david@e.co.za)              }
{                                                                              }
{ This unit is part of Delphi Fundamentals, it's original name is              }
{ cInternetStandards.                                                          }
{                                                                              }
{ I invite you to use this unit, free of charge.                               }
{ I invite you to distibute this unit, but it must be for free.                }
{ I also invite you to contribute to its development, but do not distribute    }
{ a modified copy of this file, send modifications, suggestions and bug        }
{ reports to david@e.co.za                                                     }
{                                                                              }
{                                                                              }
{ 01/02/2000  v0.01  Initial html Functions                                    }
{ 20/02/2000  v0.02  DecomposeFromField, DecodeMIME functions,                 }
{                    DecodeQuotedPrintableText, htmlSafeText                   }
{                    Document code                                             }
{ 04/03/2000  v0.03  RFC Date-Time functions                                   }
{ 05/03/2000  v0.04  HTTP response functions                                   }
{ 06/03/2000  v0.05  Message header decoding functions                         }
{ 08/03/2000  v0.06  Nested MIME messages.                                     }
{ 09/04/2000  v0.07  ThtmlParser.                                              }
{ 11/04/2000  v0.08  ThtmlRenderer.                                            }
{ 12/04/2000  v0.09  Added AhtmlRenderer and moved ThtmlRenderer to            }
{                    cHTMLRenderer.                                            }
{                    710 lines interface. 2090 lines implementation.           }
{ 15/04/2000  v0.10  Rewrote ThtmlParser from HTML 4.0 specifications.         }
{                    769 lines interface. 2400 lines implementation.           }
{ 28/04/2000  v0.11  Moved HTTP code to cHTTPServer.                           }
{                                                                              }



{                                                                              }
{ Miscellaneous                                                                }
{                                                                              }
const
  CRLF  = ASCII_CR + ASCII_LF;
  HTAB  = ASCII_HT;
  CTL   = [#0..#31];
  SPACE = [ASCII_HT, ASCII_SP];

Function TransferRate (const Length, Speed : Integer) : String;
Function HeaderField (const FieldName, FieldValue : String; const Parameters : String = '') : String;
Function RFCTime (const D : TDateTime) : String;
Function DateField : String;
Function ContentLengthField (const Length : Integer) : String;
Function ContentTypeField (const ContentType : String) : String;


{                                                                              }
{ Date/Time Encoding/Decoding                                                  }
{   From RFC822, RFC850, RFC1123, RFC1036, RFC1945 (See implementation for     }
{   details)                                                                   }
{                                                                              }
Function RFCTimeZoneToGMTBias (const Zone : String; const Strict : Boolean = False) : Integer;
{ Returns the bias of the specified Zone relative to GMT (UT) in minutes       }
Function GMTDateTimeToRFC1123DateTime (const D : TDateTime; const IncludeDayOfWeek : Boolean = True) : String;
{ Converts a GMT Date-Time to a RFC1123 Date-Time                              }
{ This is the preferred representation on the Internet for all Date-Times.     }
{ See cDateTime.LocalTimeToGMTTime for converting to GMT time.                 }
Function RFCDateTimeToGMTDateTime (const S : String) : TDateTime;
{ Converts a Date-Time that complies with any of the RFC standards to          }
{   a GMT Date-Time. This is a tolerant implementation.                        }
{ See cDateTime.GMTTimeToLocalTime for converting to local time.               }



{                                                                              }
{ Header Decoding/Encoding                                                     }
{                                                                              }
{ From RFC822 (Standard for the format of ARPA Internet Text Messages) and     }
{   RFC2616 (Hypertext Transfer Protocol -- HTTP/1.1). See implementation for  }
{   details.                                                                   }
type
  { AHeaderField                                                               }
  AHeaderField = class
    protected
    Function  GetFieldName : String; virtual; abstract;
    Procedure SetFieldName (const FieldName : String); virtual; abstract;
    Function  GetFieldBody : String; virtual; abstract;
    Procedure SetFieldBody (const FieldBody : String); virtual; abstract;
    Function  GetAsString : String; virtual;
    Procedure SetAsString (const F : String); virtual;

    public
    Constructor Create (const FieldBody : String); overload;
    Constructor Create (const FieldName, FieldBody : String); overload;

    Property AsString : String read GetAsString write SetAsString;
    Property FieldName : String read GetFieldName write SetFieldName;
    Property FieldBody : String read GetFieldBody write SetFieldBody;
  end;

  { THeaderField                                                               }
  THeaderField = class (AHeaderField)
    protected
    FFieldName : String;
    FFieldBody : String;

    Function  GetFieldName : String; override;
    Procedure SetFieldName (const FieldName : String); override;
    Function  GetFieldBody : String; override;
    Procedure SetFieldBody (const FieldBody : String); override;
  end;

  { THeader                                                                    }
  THeader = class
    protected
    HeaderFields : Array of AHeaderField;

    Procedure Decode (const Header : String); virtual;
    Function  Encode : String; virtual;

    public
    Destructor Destroy; override;

    Procedure AddField (const Field : AHeaderField); overload;
    Procedure AddField (const FieldName, FieldBody : String); overload; virtual;
    Procedure AddField (const FieldLine : String); overload;
    Function  GetField (const FieldName : String) : AHeaderField;
    Procedure SetField (const FieldName : String; const Field : AHeaderField);
    Function  HasField (const FieldName : String) : Boolean;
    Function  GetFieldBody (const FieldName : String) : String;

    Property  Field [const FieldName : String] : AHeaderField read GetField write SetField; default;
    Property  AsString : String read Encode write Decode;
  end;



{                                                                              }
{ Miscellaneous header fields                                                  }
{                                                                              }
type
  { TDateField                                                                 }
  TDateField = class (AHeaderField)
    protected
    Function  GetFieldBody : String; override;
    Procedure SetFieldBody (const FieldBody : String); override;
    Function  GetFieldName : String; override;

    public
    GMTDateTime : TDateTime;

    Constructor Create (const D : TDateTime);
  end;

  { TEMailField                                                                }
  TEMailField = class (THeaderField)
    protected
    Function  GetFieldBody : String; override;
    Procedure SetFieldBody (const FieldBody : String); override;

    public
    Address : String;
    Name    : String;
  end;



{                                                                              }
{ MIME Decoding                                                                }
{   From RFC2045                                                               }
{                                                                              }
Function MimeParameterValue (const Parameters, ParameterName : String) : String;

type
  { TMimeVersionField                                                          }
  TMimeVersionField = class (AHeaderField)
    protected
    Function  GetFieldName : String; override;
    Procedure SetFieldBody (const FieldBody : String); override;
    Function  GetFieldBody : String; override;

    public
    MajorVersion,
    MinorVersion  : Integer;
    Comment       : String;
  end;

  { TMimeContentTypeField                                                      }
  TMimeContentTypeField = class (THeaderField)
    protected
    Procedure SetFieldBody (const FieldBody : String); override;

    public
    MajorType,
    SubType     : String;
    Parameters  : String;
  end;

  { TMimeContentTransferEncodingField                                          }
  TTransferEncoding = (te_none, te_7bit, te_8bit, te_binary,
                       te_quoted_printable, te_base64, te_x_uuencode);
  TMimeContentTransferEncodingField = class (THeaderField)
    protected
    Procedure SetFieldBody (const FieldBody : String); override;

    public
    TransferEncoding : TTransferEncoding;
  end;

  { TMimeMessageHeader                                                         }
  TMimeMessageHeader = class (THeader)
    protected
    FIsMimeMessage      : Boolean;
    FBoundary           : String;
    FContentDisposition : String;
    FTransferEncoding   : TTransferEncoding;

    Procedure AddField (const FieldName, FieldBody : String); override;

    public
    Procedure Decode (const Header : String); override;
    Property  IsMimeMessage : Boolean read FIsMimeMessage;
  end;



{                                                                              }
{ Mail Message Decoding                                                        }
{   If InlineUUDecode is set TMailMessageDecoder will try to decode uuencoded  }
{     attachments contained in the message body.                               }
{                                                                              }
type
  TMessageBodyPart = class
    Header : THeader;
    Body   : String;
  end;
  TMultiPartBodyPart = class (TMessageBodyPart)
    Body : Array of TMessageBodyPart;
    Procedure AddPart (const BodyPart : TMessageBodyPart);
  end;
  TTextBodyPart = class (TMessageBodyPart)
  end;
  TAttachmentBodyPart = class (TMessageBodyPart)
    Name : String;
  end;
  TOnTextMessageLine = Function (var Line : String) : Boolean of object;
  TMailMessageDecoder = class
    protected
    FInlineUUDecode    : Boolean;
    FHeader            : THeader;
    FBody              : TMessageBodyPart;
    FOnTextMessageLine : TOnTextMessageLine;

    public
    Constructor Create;

    Property  InlineUUDecode : Boolean read FInlineUUDecode write FInlineUUDecode;
    Procedure Decode (const Header : String; const Body : TStream);
    Property  OnTextMessageLine : TOnTextMessageLine read FOnTextMessageLine write FOnTextMessageLine;

    Property  Header : THeader read FHeader write FHeader;
    Property  Body : TMessageBodyPart read FBody write FBody;
  end;





{                                                                              }
{ HTML Text Encoding                                                           }
{   From RFC 1866 (Hypertext Markup Language - 2.0)                            }
{                                                                              }
const
  { HTML "escape" encodings }
  htmlAmp            = '&amp;';        // &
  htmlLT             = '&lt;';         // <
  htmlGT             = '&gt;';         // >
  htmlQuote          = '&quot;';       // "
  htmlNBSP           = '&nbsp;';       // non-breaking space

  { HTML 3.2 "Western Latin-1 (ISO-8559-1)" character set }
  ISO_8559_1 =
        '                                 !"#$%&''()*+,-./0123456789:;<=>?' +
        '@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~'  +
        ' '  +
        '';

  { US-ASCII character set }
  US_ASCII =
        '                                 !"#$%&''()*+,-./0123456789:;<=>?' +
        '@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~'  +
        'ܢPѪ                  '  +
        '                                aGpSstFTOd8fen     vn ';


  { HTML 3.2 "Added Western Latin-1" entity set }
  htmlCopyright  = '&copy;';
  ISO_8559_1_AddedWestern : Array [1..100, 1..2] of String = (
      ('&quot;',  '&#34;'),     ('&amp;',   '&#38;'),    ('&lt;',    '&#60;'),
      ('&gt;',    '&#62;'),     ('&nbsp;',  '&#160;'),   ('&iexcl;', '&#161;'),
      ('&cent;',  '&#162;'),    ('&pound;', '&#163;'),   ('&curren;','&#164;'),
      ('&yen;',   '&#165;'),    ('&brvbar;','&#166;'),   ('&sect;',  '&#167;'),
      ('&uml;',   '&#168;'),    ('&copy;',  '&#169;'),   ('&ordf;',  '&#170;'),
      ('&laquo;', '&#171;'),    ('&not;',   '&#172;'),   ('&shy;',   '&#173;'),
      ('&reg;',   '&#174;'),    ('&macr;',  '&#175;'),   ('&deg;',   '&#176;'),
      ('&plusmn;','&#177;'),    ('&sup2;',  '&#178;'),   ('&sup3;',  '&#179;'),
      ('&acute;', '&#180;'),    ('&micro;', '&#181;'),   ('&para;',  '&#182;'),
      ('&middot;','&#183;'),    ('&cedil;', '&#184;'),   ('&sup1;',  '&#185;'),
      ('&ordm;',  '&#186;'),    ('&raquo;', '&#187;'),   ('&frac14;','&#188;'),
      ('&frac12;','&#189;'),    ('&frac34;','&#190;'),   ('&iquest;','&#191;'),
      ('&Agrave;','&#192;'),    ('&Aacute;','&#193;'),   ('&Acirc;', '&#194;'),
      ('&Atilde;','&#195;'),    ('&Auml;',  '&#196;'),   ('&Aring;', '&#197;'),
      ('&AElig;', '&#198;'),    ('&Ccedil;','&#199;'),   ('&Egrave;','&#200;'),
      ('&Eacute;','&#201;'),    ('&Ecirc;', '&#202;'),   ('&Euml;',  '&#203;'),
      ('&Igrave;','&#204;'),    ('&Iacute;','&#205;'),   ('&Icirc;', '&#206;'),
      ('&Iuml;',  '&#207;'),    ('&ETH;',   '&#208;'),   ('&Ntilde;','&#209;'),
      ('&Ograve;','&#210;'),    ('&Oacute;','&#211;'),   ('&Ocirc;', '&#212;'),
      ('&Otilde;','&#213;'),    ('&Ouml;',  '&#214;'),   ('&times;', '&#215;'),
      ('&Oslash;','&#216;'),    ('&Ugrave;','&#217;'),   ('&Uacute;','&#218;'),
      ('&Ucirc;', '&#219;'),    ('&Uuml;',  '&#220;'),   ('&Yacute;','&#221;'),
      ('&THORN;', '&#222;'),    ('&szlig;', '&#223;'),   ('&agrave;','&#224;'),
      ('&aacute;','&#225;'),    ('&acirc;', '&#226;'),   ('&atilde;','&#227;'),
      ('&auml;',  '&#228;'),    ('&aring;', '&#229;'),   ('&aelig;', '&#230;'),
      ('&ccedil;','&#231;'),    ('&egrave;','&#232;'),   ('&eacute;','&#233;'),
      ('&ecirc;', '&#234;'),    ('&euml;',  '&#235;'),   ('&igrave;','&#236;'),
      ('&iacute;','&#237;'),    ('&icirc;', '&#238;'),   ('&iuml;',  '&#239;'),
      ('&eth;',   '&#240;'),    ('&ntilde;','&#241;'),   ('&ograve;','&#242;'),
      ('&oacute;','&#243;'),    ('&ocirc;', '&#244;'),   ('&otilde;','&#245;'),
      ('&ouml;',  '&#246;'),    ('&divide;','&#247;'),   ('&oslash;','&#248;'),
      ('&ugrave;','&#249;'),    ('&uacute;','&#250;'),   ('&ucirc;', '&#251;'),
      ('&uuml;',  '&#252;'),    ('&yacute;','&#253;'),   ('&thorn;', '&#254;'),
      ('&yuml;',  '&#255;'));

  { HTML elements }
  htmlBR             = '<BR>' + CRLF;
  htmlHorisontalRule = '<HR>' + CRLF;


{ Text encoding                                                                }
Function htmlSafeText (const Txt : String) : String;
Function htmlChar (const Value : Byte) : String;

{ HTML Document                                                                }
Function htmlComment (const Comment : String) : String;
Function htmlDocument (const Title, Txt : String;
         const Description : String = '';
         const Keywords : String = '';
         const BodyBackColour : String = '';
         const BodyTextColour : String = '';
         const BodyLinkColour : String = '') : String;

{ Text format                                                                  }
Function htmlPreformattedText (const Txt : String; const WidthChars : Integer = 0) : String;
Function htmlBold (const Txt : String) : String;
Function htmlItalic (const Txt : String) : String;
Function htmlUnderline (const Txt : String) : String;
Function htmlStrikethrough (const Txt : String) : String;
Function htmlTeletypeFont (const Txt : String) : String;
Function htmlFont (const Txt : String; const FontSize : Integer = 0;
         const FontColour : String = ''; const FontFace : String = '') : String;

{ Formatting                                                                   }
Function htmlCenter (const Txt : String) : String;

{ Tables                                                                       }
Function htmlTable (const Txt : String; const Border : Boolean = False;
         const Width : String = ''; const BackColour : String = '';
         const CellSpacing : Integer = 2; const CellPadding : Integer = 1) : String;
Function htmlBorderedTable (const Txt : String; const Width : String = '';
         const FrameWidth : Integer = 2; const FrameColour : String = '';
         const CellPadding : Integer = 1) : String;
Function htmlRow (const Txt : String; const BackColour : String = '') : String;
Function htmlCol (const Txt : String; const ColSpan : Integer = 1;
         const Width : String = ''; const BackColour : String = '') : String;

{ Other Tags                                                                   }
Function htmlImage (const Img : String; const Align : String = '';
         const AltText : String = '';
         const Width : Integer = 0; const Height : Integer = 0;
         const Border : Boolean = True) : String;
Function htmlLink (const Link : String; const Txt : String) : String;



{                                                                              }
{ HTML Graphics Rendering                                                      }
{                                                                              }
{ AhtmlRenderer                                                                }
{   Base class for a visual HTML renderer.                                     }
{   HorSpace returns the numbers of pixels left horisontally before the        }
{     Renderer will attempt a Break.                                           }
{   BOL returns True if renderer is at the beginning of a line.                }
{   AddControl gives a visual control to a renderer to be added to the         }
{     display.                                                                 }
{   BreakLeft moves to beginning of the next rendering "line" (newline). An    }
{     additional SkipHeight pixel lines are skipped.                           }
{                                                                              }
type
  ThtmlRenderAttributes = Set of (raBold, raItalic, raUnderline, raSubscript,
                                  raSuperScript, raStrikethrough,
                                  raStrong, raEmphasis, raCode, raKeyboard,
                                  raVariable, raTeletype, raCitation,
                                  raDefinition, raSample, raPreformatted,
                                  raCentered, raSmall, raBig, raIsLink);
type
  ThtmlFontSize = Integer;
  ThtmlFontColour = String;
  ThtmlRendererBreakType = (btMargin,
                            btNextMargin,
                            btLeftMargin);
  ThtmlHorAlign = (haLeft, haRight, haCenter);
  ThtmlVerAlign = (vaBottom, vaTop, vaCenter);
  AhtmlRenderer = class
    protected
    FFontSize       : ThtmlFontSize;
    FFontColour     : ThtmlFontColour;
    FFontBackColour : ThtmlFontColour;
    FAttributes     : ThtmlRenderAttributes;

    FIndent         : Integer;
    FDesiredWidth   : Integer;
    FVAlign         : ThtmlVerAlign;
    FHAlign         : ThtmlHorAlign;

    public
    Constructor Create (const Parent : AhtmlRenderer);

    { HTML attributes }
    Property  FontSize : ThtmlFontSize read FFontSize write FFontSize;
    Property  FontColour : ThtmlFontColour read FFontColour write FFontColour;
    Property  FontBackColour : ThtmlFontColour read FFontBackColour write FFontBackColour;
    Property  Attributes : ThtmlRenderAttributes read FAttributes write FAttributes;

    { Rendering attributes }
    Property  DesiredWidth : Integer read FDesiredWidth write FDesiredWidth;
    Property  Indent : Integer read FIndent write FIndent;
    Function  HorSpace : Integer; virtual; abstract;
    Function  BOL : Boolean; virtual; abstract;
    Function  LeftMargin : Boolean; virtual; abstract;
    Property  VerAlign : ThtmlVerAlign read FVAlign write FVAlign;
    Property  HorAlign : ThtmlHorAlign read FHAlign write FHAlign;

    { Rendering methods }
    Procedure BreakLine (const BreakType : ThtmlRendererBreakType = btMargin; const SkipHeight : Integer = 0); virtual; abstract;
    Procedure AddControl (const Control : TControl; const SetMargin : Boolean = False); virtual; abstract;
  end;



{                                                                              }
{ HTML Representations                                                         }
{   Base classes to represent components of a HTML document.                   }
{                                                                              }
type
  { AhtmlComponent                                                             }
  {   Base class for all entities in the HTML parse tree.                      }
  {   GetAsHTML returns the HTML representation of the component.              }
  {   Render renders the component on the Renderer.                            }
  AhtmlComponent = class
    Function  GetAsHTML : String; virtual; abstract;
    Procedure Render (const Renderer : AhtmlRenderer); virtual; abstract;
  end;
  ThtmlComponent = class (AhtmlComponent)
    Procedure Render (const Renderer : AhtmlRenderer); override;
  end;

  { ThtmlText                                                                  }
  {   Represents the content (text) of HTML.                                   }
  ThtmlText = class (AhtmlComponent)
    protected
    FText : String;

    public
    Constructor Create (const Text : String);
    Function GetAsHTML : String; override;
  end;

  { ThtmlEntity                                                                }
  {   Represents a HTML entity (&<element>)                                    }
  ThtmlEntity = class (AhtmlComponent)
    protected
    FEntity : String;

    public
    Constructor Create (const Entity : String);
    Function GetAsHTML : String; override;
  end;

  { AhtmlElement                                                               }
  {   Represents a HTML element (<tag attr>)                                   }
  AhtmlElement = class (AhtmlComponent)
    protected
    Function GetName : String; virtual; abstract;
    Function GetCloseElement : Boolean; virtual; abstract;

    public
    Property Name : String read GetName;
    Property CloseElement : Boolean read GetCloseElement;
    Function AttrAsString (const Attr : String; const DefaultValue : String = '') : String; virtual; abstract;
    Function AttrAsInteger (const Attr : String; const DefaultValue : Integer = -1) : Integer; virtual; abstract;
  end;
  ThtmlAttribute = record
    Attr  : String;
    Value : String;
  end;
  ThtmlAttributes = class
    protected
    FAttr : Array of ThtmlAttribute;

    Function IndexOf (const Attr : String) : Integer;

    public
    Procedure AddAttribute (const Attr : String; const Value : String = '');
    Function  GetAsHTML : String;
    Function  AsString (const Attr : String; const DefaultValue : String = '') : String;
    Function  AsInteger (const Attr : String; const DefaultValue : Integer = -1) : Integer;
    Function  Duplicate : ThtmlAttributes;
  end;
  ThtmlElement = class (AhtmlElement)
    protected
    FName : String;
    FAttr : ThtmlAttributes;

    Function GetName : String; override;
    Function GetCloseElement : Boolean; override;

    public
    Constructor Create (const Name : String; const Attributes : ThtmlAttributes = nil);
    Destructor  Destroy; override;

    Function  GetAsHTML : String; override;
    Property  Name : String read GetName;
    Property  CloseElement : Boolean read GetCloseElement;
    Function  AttrAsString (const Attr : String; const DefaultValue : String = '') : String; override;
    Function  AttrAsInteger (const Attr : String; const DefaultValue : Integer = -1) : Integer; override;
    Procedure Render (const Renderer : AhtmlRenderer); override;
  end;

  { AhtmlTag                                                                   }
  {   Represents a HTML tag (<open element> <content> <close element>)         }
  AhtmlTag = class (AhtmlComponent)
    protected
    Function GetEnclosed : AhtmlComponent; virtual; abstract;

    public
    Property Enclosed : AhtmlComponent read GetEnclosed;
  end;
  ThtmlTag = class (AhtmlTag)
    protected
    FOpenElement,
    FCloseElement : AhtmlElement;
    FEnclosed     : AhtmlComponent;

    Function GetEnclosed : AhtmlComponent; override;

    public
    Constructor Create (const OpenElement, CloseElement : AhtmlElement;
                        const Enclosed : AhtmlComponent);
    Destructor Destroy; override;

    Function  GetAsHTML : String; override;
    Procedure Render (const Renderer : AhtmlRenderer); override;
  end;

  { ThtmlConcatenation                                                         }
  {   Represents an orderered list of HTML components.                         }
  ThtmlConcatenation = class (AhtmlComponent)
    protected
    FComponents : Array of AhtmlComponent;

    Function GetComponent (const Idx : Integer) : AhtmlComponent;

    public
    Constructor Create (const Left, Right : AhtmlComponent);
    Destructor Destroy; override;

    Function  GetAsHTML : String; override;
    Procedure Render (const Renderer : AhtmlRenderer); override;

    Procedure Add (const C : AhtmlComponent);
    Function  Count : Integer;
    Property  Component [const Idx : Integer] : AhtmlComponent read GetComponent;
  end;

  { ThtmlASPCode                                                               }
  {   Represents ASP (Active Server Pages) code.                               }
  ThtmlASPCode = class (ThtmlComponent)
    protected
    FCode : String;

    public
    Constructor Create (const Code : String);
    Function GetAsHTML : String; override;
  end;

  { ThtmlComment                                                               }
  ThtmlComment = class (ThtmlComponent)
    protected
    FComment : String;

    public
    Constructor Create (const Comment : String);
    Function GetAsHTML : String; override;
  end;

  { ThtmlDTDTag                                                                }
  {   Representation of a DTD (document type declaration)                      }
  ThtmlDTDTag = class (ThtmlComponent)
    protected
    FText : String;

    public
    Constructor Create (const Text : String);
    Function GetAsHTML : String; override;
  end;

  { ThtmlProcessingInstructions                                                }
  ThtmlProcessingInstructions = class (ThtmlComponent)
    protected
    FPI : String;

    public
    Constructor Create (const PI : String);
    Function GetAsHTML : String; override;
  end;

  { ThtmlSSITag                                                                }
  {   Represents an SSI (Server Side Include) tag.                             }
  ThtmlSSITag = class (ThtmlComponent)
    protected
    FText : String;

    public
    Constructor Create (const Text : String);
    Function GetAsHTML : String; override;
  end;

  { ThtmlUnknownElement                                                        }
  {   Represents unknown elements.                                             }
  ThtmlUnknownElement = class (ThtmlComponent)
    protected
    FElement : String;

    public
    Constructor Create (const Element : String);
    Function GetAsHTML : String; override;
  end;



{                                                                              }
{ HTML Decoding                                                                }
{   ParseHTMLTree returns a tree of AhtmlComponents representing the decoded   }
{     HTML text. If an application wish to insert custom AhtmlComponent        }
{     classes into the tree it should override the Interpreted... functions.   }
{                                                                              }
type
  EhtmlParser = class (Exception);
  ThtmlParser = class
    protected
    FToken  : AhtmlComponent;
    FS      : TStream;

    Function GetNextToken : AhtmlComponent;

    Function InterpretedTag (const Element : AhtmlElement) : AhtmlElement; overload; virtual;
    Function InterpretedTag (const OpenElement, CloseElement : AhtmlElement; const Enclosed : AhtmlComponent) : AhtmlTag; overload; virtual;
    Function InterpretedText (const Text : String) : AhtmlComponent; virtual;
    Function InterpretedEntity (const Entity : String) : AhtmlComponent; virtual;

    public
    Function ParseHTMLTree (const S : TStream) : AhtmlComponent;
    Function ParseHTMLFlat (const S : TStream) : AhtmlComponent;
  end;

Function  htmlColour (const Colour : String) : TColor;
{ Returns the Delphi TColor given an HTML Colour string.                       }

type
  ThtmlSizeType = (stNone, stPixels, stPercent, stRelative);

Procedure htmlDecodeSize (const S : String; var Size : Integer; var SizeType : ThtmlSizeType);
{ Decodes a HTML size (pixels, percentage or relative).                        }

{ Translate                                                                    }
{   Default Translate function translates from the ISO-8559-1 character set.   }
Function Translate (const Ch : Char; const TranslationTable : String = ISO_8559_1) : Char; overload;
Function Translate (const S : String; const TranslationTable : String = ISO_8559_1) : String; overload;



implementation



uses
  // L0 units
  cUtil,
  cDateTime,  // LocalTimeToGMTTime
  cMaths;     // HexToInt


{                                                                              }
{ Miscellaneous                                                                }
{                                                                              }
Function TransferRate (const Length, Speed : Integer) : String;
var cps : Double;
  Begin
    if Speed <= 0 then
      Result := '' else
      begin
        cps := Length / (Speed / 1000.0);
        if cps >= 1024 * 1024 then
          Result := Format ('%0.1f', [cps / (1024 * 1024)]) + 'Mb/s' else
        if cps >= 1024 then
          Result := Format ('%0.1f', [cps / 1024]) + 'Kb/s' else
          Result := IntToStr (Round (cps)) + 'cps';
      end;
  End;

Function HeaderField (const FieldName, FieldValue : String; const Parameters : String = '') : String;
  Begin
    Result := FieldName + ': ' + FieldValue +
              Cond (Parameters <> '', '; ' + Parameters, '') +
              CRLF;
  End;

Function RFCTime (const D : TDateTime) : String;
  Begin
    Result := GMTDateTimeToRFC1123DateTime (LocalTimeToGMTTime (D));
  End;

Function DateField : String;
  Begin
    Result := HeaderField ('Date', GMTDateTimeToRFC1123DateTime (LocalTimeToGMTTime (Now)));
  End;

Function ContentLengthField (const Length : Integer) : String;
  Begin
    Result := HeaderField ('Content-Length', IntToStr (Length));
  End;

Function ContentTypeField (const ContentType : String) : String;
  Begin
    Result := HeaderField ('Content-Type', ContentType);
  End;



{                                                                              }
{ Date/Time Encoding/Decoding                                                  }
{                                                                              }
{ From RFC 822 (Standard for the format of ARPA INTERNET Text Messages):       }
{    "time        =  hour zone                      ; ANSI and Military        }
{     hour        =  2DIGIT ":" 2DIGIT [":" 2DIGIT] ; 00:00:00 - 23:59:59      }
{     zone        =  "UT"  / "GMT"                  ; Universal Time           }
{                                                   ; North American : UT      }
{                 /  "EST" / "EDT"                  ;  Eastern:  - 5/ - 4      }
{                 /  "CST" / "CDT"                  ;  Central:  - 6/ - 5      }
{                 /  "MST" / "MDT"                  ;  Mountain: - 7/ - 6      }
{                 /  "PST" / "PDT"                  ;  Pacific:  - 8/ - 7      }
{                 /  1ALPHA                         ; Military: Z = UT;        }
{                                                   ;  A:-1; (J not used)      }
{                                                   ;  M:-12; N:+1; Y:+12      }
{                 / ( ("+" / "-") 4DIGIT )          ; Local differential       }
{                                                   ;  hours+min. (HHMM)       }
{     date-time   =  [ day "," ] date time          ; dd mm yy                 }
{                                                   ;  hh:mm:ss zzz            }
{     day         =  "Mon"  / "Tue" /  "Wed"  / "Thu"                          }
{                 /  "Fri"  / "Sat" /  "Sun"                                   }
{     date        =  1*2DIGIT month 2DIGIT        ; day month year             }
{                                                 ;  e.g. 20 Jun 82            }
{     month       =  "Jan"  /  "Feb" /  "Mar"  /  "Apr"                        }
{                 /  "May"  /  "Jun" /  "Jul"  /  "Aug"                        }
{                 /  "Sep"  /  "Oct" /  "Nov"  /  "Dec"                    "   }
{                                                                              }
{ Note that even though RFC 822 states hour=2DIGIT":"2DIGIT, none of the       }
{   examples given in the appendix include the ":",                            }
{   for example: "26 Aug 76 1429 EDT"                                          }
{                                                                              }
{                                                                              }
{ From RFC 1036 (Standard for Interchange of USENET Messages):                 }
{                                                                              }
{   "Its format must be acceptable both in RFC-822 and to the getdate(3)       }
{    routine that is provided with the Usenet software.   ...                  }
{    One format that is acceptable to both is:                                 }
{                                                                              }
{                      Wdy, DD Mon YY HH:MM:SS TIMEZONE                        }
{                                                                              }
{    Note in particular that ctime(3) format:                                  }
{                                                                              }
{                          Wdy Mon DD HH:MM:SS YYYY                            }
{                                                                              }
{    is not acceptable because it is not a valid RFC-822 date.  However,       }
{    since older software still generates this format, news                    }
{    implementations are encouraged to accept this format and translate        }
{    it into an acceptable format.                                         "   }
{                                                                              }
{   "Here is an example of a message in the old format (before the             }
{    existence of this standard). It is recommended that                       }
{    implementations also accept messages in this format to ease upward        }
{    conversion.                                                               }
{                                                                              }
{               Posted: Fri Nov 19 16:14:55 1982                           "   }
{                                                                              }
{                                                                              }
{ From RFC 1945 (Hypertext Transfer Protocol -- HTTP/1.0)                      }
{                                                                              }
{  "HTTP/1.0 applications have historically allowed three different            }
{   formats for the representation of date/time stamps:                        }
{                                                                              }
{       Sun, 06 Nov 1994 08:49:37 GMT    ; RFC 822, updated by RFC 1123        }
{       Sunday, 06-Nov-94 08:49:37 GMT   ; RFC 850, obsoleted by RFC 1036      }
{       Sun Nov  6 08:49:37 1994         ; ANSI C's asctime() format           }
{                                                                              }
{   The first format is preferred as an Internet standard and represents       }
{   a fixed-length subset of that defined by RFC 1123 [6] (an update to        }
{   RFC 822 [7]). The second format is in common use, but is based on the      }
{   obsolete RFC 850 [10] date format and lacks a four-digit year.             }
{   HTTP/1.0 clients and servers that parse the date value should accept       }
{   all three formats, though they must never generate the third               }
{   (asctime) format.                                                          }
{                                                                              }
{      Note: Recipients of date values are encouraged to be robust in          }
{      accepting date values that may have been generated by non-HTTP          }
{      applications, as is sometimes the case when retrieving or posting       }
{      messages via proxies/gateways to SMTP or NNTP.                       "  }
{                                                                              }
{  "All HTTP/1.0 date/time stamps must be represented in Universal Time        }
{   (UT), also known as Greenwich Mean Time (GMT), without exception.          }
{                                                                              }
{       HTTP-date      = rfc1123-date | rfc850-date | asctime-date             }
{                                                                              }
{       rfc1123-date   = wkday "," SP date1 SP time SP "GMT"                   }
{       rfc850-date    = weekday "," SP date2 SP time SP "GMT"                 }
{       asctime-date   = wkday SP date3 SP time SP 4DIGIT                      }
{                                                                              }
{       date1          = 2DIGIT SP month SP 4DIGIT                             }
{                        ; day month year (e.g., 02 Jun 1982)                  }
{       date2          = 2DIGIT "-" month "-" 2DIGIT                           }
{                        ; day-month-year (e.g., 02-Jun-82)                    }
{       date3          = month SP ( 2DIGIT | ( SP 1DIGIT ))                    }
{                        ; month day (e.g., Jun  2)                            }
{                                                                              }
{       time           = 2DIGIT ":" 2DIGIT ":" 2DIGIT                          }
{                        ; 00:00:00 - 23:59:59                                 }
{                                                                              }
{       wkday          = "Mon" | "Tue" | "Wed"                                 }
{                      | "Thu" | "Fri" | "Sat" | "Sun"                         }
{                                                                              }
{       weekday        = "Monday" | "Tuesday" | "Wednesday"                    }
{                      | "Thursday" | "Friday" | "Saturday" | "Sunday"         }
{                                                                              }
{       month          = "Jan" | "Feb" | "Mar" | "Apr"                         }
{                      | "May" | "Jun" | "Jul" | "Aug"                         }
{                      | "Sep" | "Oct" | "Nov" | "Dec"                      "  }
{                                                                              }
Function RFCTimeZoneToGMTBias (const Zone : String; const Strict : Boolean = False) : Integer;
type TZoneBias = record
       Zone : String;
       Bias : Integer;
     end;
const TimeZones = 35;
      ZoneBias : Array [1..TimeZones] of TZoneBias =
          ((Zone:'GMT'; Bias:0),                 (Zone:'UT';  Bias:0),
           (Zone:'EST'; Bias:-5*60),             (Zone:'EDT'; Bias:-4*60),
           (Zone:'CST'; Bias:-6*60),             (Zone:'CDT'; Bias:-5*60),
           (Zone:'MST'; Bias:-7*60),             (Zone:'MDT'; Bias:-6*60),
           (Zone:'PST'; Bias:-8*60),             (Zone:'PDT'; Bias:-7*60),
           (Zone:'Z';   Bias:0),                 (Zone:'A';   Bias:-1*60),
           (Zone:'B';   Bias:-2*60),             (Zone:'C';   Bias:-3*60),
           (Zone:'D';   Bias:-4*60),             (Zone:'E';   Bias:-5*60),
           (Zone:'F';   Bias:-6*60),             (Zone:'G';   Bias:-7*60),
           (Zone:'H';   Bias:-8*60),             (Zone:'I';   Bias:-9*60),
           (Zone:'K';   Bias:-10*60),            (Zone:'L';   Bias:-11*60),
           (Zone:'M';   Bias:-12*60),            (Zone:'N';   Bias:1*60),
           (Zone:'O';   Bias:2*60),              (Zone:'P';   Bias:3*60),
           (Zone:'Q';   Bias:4*60),              (Zone:'R';   Bias:3*60),
           (Zone:'S';   Bias:6*60),              (Zone:'T';   Bias:3*60),
           (Zone:'U';   Bias:8*60),              (Zone:'V';   Bias:3*60),
           (Zone:'W';   Bias:10*60),             (Zone:'X';   Bias:3*60),
           (Zone:'Y';   Bias:12*60));
var S : String;
    I : Integer;
  Begin
    if Zone [1] in ['+', '-'] then // +hhmm format
      begin
        if Strict and (Length (Zone) <> 5) then
          raise EConvertError.Create ('Not a RFC1123 time zone: Invalid differential');
        S := Trim (Zone, SPACE);
        try
          I := StrToInt (CopyLeft (S, 3));
        except
          raise EConvertError.Create ('Not a RFC1123 time zone: Invalid hour in differential');
        end;
        if Abs (I) > 12 then
          raise EConvertError.Create ('Not a RFC1123 time zone: Invalid hour in differential');
        Result := I * 60;
        S := CopyFrom (S, 4);
        if S <> '' then
          if Length (S) <> 2 then
            raise EConvertError.Create ('Not a RFC1123 time zone: Invalid minute in differential') else
            try
              I := StrToInt (S);
              Result := Result + I;
            except
              raise EConvertError.Create ('Not a RFC1123 time zone: Invalid minute in differential');
            end;
      end else
      begin // named format
        S := UpperCase (Zone);
        For I := 1 to TimeZones do
          if ZoneBias [I].Zone = S then
            begin
              Result := ZoneBias [I].Bias;
              exit;
            end;
        raise EConvertError.Create ('Not a RFC1123 time zone: Unknown zone name');
      end;
  End;

{ Converts a RFC Time to GMT Time                                              }
Function RFCTimeToGMTTime (const S : String; const StrictlyRFC1123 : Boolean = False) : TDateTime;
var I : Integer;
    T : String;
    HH, MM, SS : Integer;
    U : StringArray;
  Begin
    T := Trim (S, SPACE);

    // Get Zone bias
    I := PosPrev (SPACE, T);
    if I > 0 then
      begin
        Result := Int (RFCTimeZoneToGMTBias (CopyFrom (T, I + 1), StrictlyRFC1123)) / 1440.0;
        T := Trim (CopyLeft (T, I - 1), SPACE);
      end else
      if StrictlyRFC1123 then
        raise EConvertError.Create ('Not a RFC1123 time: No Zone') else
        Result := 0;

    // Get time
    U := Split (T, ':');
    if not StrictlyRFC1123 and (Length (U) = 1) and (Length (U [0]) = 4) then
      try // old hhmm format
        HH := StrToInt (Copy (U [0], 1, 2));
        MM := StrToInt (Copy (U [0], 3, 2));
        SS := 0;
      except
        raise EConvertError.Create ('Not a RFC time: Invalid time');
      end else
      begin // hh:mm[:ss] format (RFC1123)
        if (Length (U) < 2) or (Length (U) > 3) then
          raise EConvertError.Create ('Not a RFC1123 time: Invalid time');
        try
          HH := StrToInt (Trim (U [0], SPACE));
          MM := StrToInt (Trim (U [1], SPACE));
          if Length (U) = 3 then
            SS := StrToInt (Trim (U [2], SPACE)) else
            SS := 0;
        except
          raise EConvertError.Create ('Not a RFC1123 time: Invalid time');
        end;
      end;

    if (HH < 0) or (HH > 23) then
      raise EConvertError.Create ('Not a RFC1123 time: Invalid hour');
    if (MM < 0) or (MM > 59) then
      raise EConvertError.Create ('Not a RFC1123 time: Invalid minute');
    if (SS < 0) or (SS > 59) then
      raise EConvertError.Create ('Not a RFC1123 time: Invalid second');
    Result := Result + EncodeTime (HH, MM, SS, 0);
  End;

{ Converts GMT Time to a RFC1123 Time                                          }
Function GMTTimeToRFC1123Time (const D : TDateTime; const IncludeSeconds : Boolean = True) : String;
var Ho, Mi, Se, Ms : Word;
  Begin
    DecodeTime (D, Ho, Mi, Se, Ms);
    Result := PadLeft (IntToStr (Ho), '0', 2) + ':' +
              PadLeft (IntToStr (Mi), '0', 2) +
              Cond (IncludeSeconds, ':' + PadLeft (IntToStr (Se), '0', 2), '') +
              ' GMT';
  End;


{ Date constants                                                               }
const
  RFC850DayNames : Array [1..7] of String = (
      'Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday');
  RFC1123DayNames : Array [1..7] of String = (
      'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat');
  RFCMonthNames : Array [1..12] of String = (
      'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');

Function RFC850DayOfWeek (const S : String) : Integer;
var I : Integer;
    T : String;
  Begin
    T := UpperCase (S);
    For I := 1 to 7 do
      if UpperCase (RFC850DayNames [I]) = T then
        begin
          Result := I;
          exit;
        end;
    Result := -1;
  End;

Function RFC1123DayOfWeek (const S : String) : Integer;
var I : Integer;
    T : String;
  Begin
    T := UpperCase (S);
    For I := 1 to 7 do
      if UpperCase (RFC1123DayNames [I]) = T then
        begin
          Result := I;
          exit;
        end;
    Result := -1;
  End;

Function RFCMonth (const S : String) : Integer;
var I : Integer;
    T : String;
  Begin
    T := UpperCase (S);
    For I := 1 to 12 do
      if UpperCase (RFCMonthNames [I]) = T then
        begin
          Result := I;
          exit;
        end;
    Result := -1;
  End;

Function GMTDateTimeToRFC1123DateTime (const D : TDateTime; const IncludeDayOfWeek : Boolean = True) : String;
var Ye, Mo, Da : Word;
  Begin
    DecodeDate (D, Ye, Mo, Da);
    Result := Cond (IncludeDayOfWeek, RFC1123DayNames [DayOfWeek (D)] + ', ', '') +
              PadLeft (IntToStr (Da), '0', 2) + ' ' +
              RFCMonthNames [Mo] + ' ' +
              IntToStr (Ye) + ' ' +
              GMTTimeToRFC1123Time (D);
  End;

Function RFCDateTimeToGMTDateTime (const S : String) : TDateTime;
var T, U : String;
    I : Integer;
    D, M, Y, DOW : Integer;
    V, W : StringArray;
  Begin
    T := Trim (S, SPACE);

    // Extract Day of week
    I := PosNext (SPACE + [','], T);
    if I = 0 then
      DOW := -1 else
      begin
        U := CopyLeft (T, I - 1);
        DOW := RFC850DayOfWeek (U);
        if DOW = -1 then
          DOW := RFC1123DayOfWeek (U);
        if DOW <> -1 then
          T := Trim (CopyFrom (S, I + 1), SPACE);
      end;

    V := Split (T, SPACE);
    if Length (V) < 3 then
      raise EConvertError.Create ('Not a recognised date: Too few elements');

    if PosNext ('-', V [0]) > 0 then // RFC850 date, eg "Sunday, 06-Nov-94 08:49:37 GMT"
      begin
        W := Split (V [0], '-');
        if Length (W) <> 3 then
          raise EConvertError.Create ('Not a recognised date: Not a RFC850 date');
        M := RFCMonth (W [1]);
        if M = -1 then
          raise EConvertError.Create ('Not a recognised date: Not a RFC850 date');
        try
          D := StrToInt (W [0]);
          Y := StrToInt (W [2]);
          if Y < 100 then
            Y := TwoDigitYearToYear (Y);
        except
          raise EConvertError.Create ('Not a recognised date: Not a RFC850 date');
        end;
        Result := EncodeDate (Y, M, D) + RFCTimeToGMTTime (V [1]  + V [2]);
        exit;
      end;

    M := RFCMonth (V [1]);
    if M >= 1 then // RFC822 date, eg Sun, 06 Nov 1994 08:49:37 GMT
      begin
        try
          D := StrToInt (V [0]);
          Y := StrToInt (V [2]);
        except
          raise EConvertError.Create ('Not a recognised date: Not a RFC822 date');
        end;
        Result := EncodeDate (Y, M, D);
        if Length (V) = 4 then
          Result := Result + RFCTimeToGMTTime (V [3]) else
          if Length (V) >= 5 then
            Result := Result + RFCTimeToGMTTime (V [3] + ' ' + V [4]);
        exit;
      end;

    M := RFCMonth (V [0]);
    if M >= 1 then // ANSI C asctime() format, eg "Sun Nov  6 08:49:37 1994"
      begin
        try
          D := StrToInt (V [1]);
          Y := StrToInt (V [3]);
        except
          raise EConvertError.Create ('Not a recognised date: Not a asctime() date');
        end;
        Result := EncodeDate (Y, M, D) + RFCTimeToGMTTime (V [2]);
      end else
      raise EConvertError.Create ('Not a recognised date.');
  End;



{                                                                              }
{ Header Decoding/Encoding                                                     }
{                                                                              }
{ From RFC822 (Standard for the format of ARPA Internet Text Messages):        }
{     field       =  field-name ":" [ field-body ] CRLF                        }
{     field-name  =  1*<any CHAR, excluding CTLs, SPACE, and ":">              }
{     field-body  =  field-body-contents                                       }
{                    [CRLF LWSP-char field-body]                               }
{     LWSP-char   =  SPACE / HTAB                 ; semantics = SPACE          }
{                                                                              }
{ From RFC2616 (Hypertext Transfer Protocol -- HTTP/1.1):                      }
{   HTTP/1.1 header field values can be folded onto multiple lines if the      }
{   continuation line begins with a space or horizontal tab. All linear        }
{   white space, including folding, has the same semantics as SP. A            }
{   recipient MAY replace any linear white space with a single SP before       }
{   interpreting the field value or forwarding the message downstream.         }
{       LWS            = [CRLF] 1*( SP | HT )                                  }

{ AHeaderField                                                                 }
Constructor AHeaderField.Create (const FieldBody : String);
  Begin
    inherited Create;
    SetFieldBody (FieldBody);
  End;

Constructor AHeaderField.Create (const FieldName, FieldBody : String);
  Begin
    inherited Create;
    SetFieldName (FieldName);
    SetFieldBody (FieldBody);
  End;

Function AHeaderField.GetAsString : String;
  Begin
    Result := FieldName + ': ' + FieldBody + CRLF;
  End;

Procedure AHeaderField.SetAsString (const F : String);
var I : Integer;
    S : String;
  Begin
    S := Trim (F, SPACE);
    I := PosNext (':', F);
    if I <= 1 then
      raise EConvertError.Create ('Not a valid header field');
    FieldName := CopyLeft (S, I - 1);
    FieldBody := Trim (CopyFrom (S, I + 1), SPACE);
  End;

{ THeaderField                                                                 }
Function THeaderField.GetFieldName : String;
  Begin
    Result := FFieldName;
  End;

Procedure THeaderField.SetFieldName (const FieldName : String);
  Begin
    FFieldName := FieldName;
  End;

Function THeaderField.GetFieldBody : String;
  Begin
    Result := FFieldBody;
  End;

Procedure THeaderField.SetFieldBody (const FieldBody : String);
  Begin
    FFieldBody := FieldBody;
  End;

{ THeader                                                                      }
Destructor THeader.Destroy;
var I : Integer;
  Begin
    For I := Length (HeaderFields) - 1 downto 0 do
      HeaderFields [I].Free;
    SetLength (HeaderFields, 0);
    inherited Destroy;
  End;

Procedure THeader.AddField (const Field : AHeaderField);
var L : Integer;
  Begin
    L := Length (HeaderFields);
    SetLength (HeaderFields, L + 1);
    HeaderFields [L] := Field;
  End;

Procedure THeader.AddField (const FieldName, FieldBody : String);
var F : THeaderField;
  Begin
    F := THeaderField.Create;
    F.FieldName := FieldName;
    F.FieldBody := FieldBody;
    AddField (F);
  End;

Procedure THeader.AddField (const FieldLine : String);
var I : Integer;
  Begin
    I := PosNext (':', FieldLine);
    if I = 0 then
      raise EConvertError.Create ('Not a valid field line');
    AddField (Trim (CopyLeft (FieldLine, I - 1)), Trim (CopyFrom (FieldLine, I + 1)));
  End;

Function THeader.GetField (const FieldName : String) : AHeaderField;
var I : Integer;
    S : String;
  Begin
    S := UpperCase (Trim (FieldName, SPACE));
    For I := 0 to Length (HeaderFields) - 1 do
      if S = UpperCase (HeaderFields [I].FieldName) then
        begin
          Result := HeaderFields [I];
          exit;
        end;
    Result := nil;
  End;

Procedure THeader.SetField (const FieldName : String; const Field : AHeaderField);

  Procedure DeleteField (const FieldIndex : Integer);
  var L : Integer;
    Begin
      L := Length (HeaderFields);
      if FieldIndex + 1 < L then
        Move (HeaderFields [FieldIndex + 1], HeaderFields [FieldIndex], Sizeof (AHeaderField) * (L - FieldIndex - 1));
      SetLength (HeaderFields, L - 1);
    End;

var I : Integer;
    S : String;
  Begin
    S := UpperCase (Trim (FieldName, SPACE));
    For I := 0 to Length (HeaderFields) - 1 do
      if S = UpperCase (HeaderFields [I].FieldName) then
        begin
          FreeAndNil (HeaderFields [I]);
          if Assigned (Field) then
            HeaderFields [I] := Field else
            DeleteField (I);
          exit;
        end;
    if Assigned (Field) then
      AddField (Field);
  End;

Function THeader.HasField (const FieldName : String) : Boolean;
  Begin
    Result := Assigned (GetField (FieldName));
  End;

Function THeader.GetFieldBody (const FieldName : String) : String;
var F : AHeaderField;
  Begin
    F := GetField (FieldName);
    if Assigned (F) then
      Result := F.FieldBody else
      Result := '';
  End;

Function THeader.Encode : String;
var I : Integer;
  Begin
    Result := '';
    For I := 0 to Length (HeaderFields) - 1 do
      Result := Result + HeaderFields [I].AsString;
    Result := Result + CRLF;
  End;

Procedure THeader.Decode (const Header : String);
var S     : StringArray;
    I, J  : Integer;
    Name,
    Body  : String;
  Begin
    S := Split (Header, CRLF);
    Name := '';
    For I := 0 to Length (S) - 1 do
      begin
        J := PosNext (':', S [I]);
        if J > 0 then // header-field
          begin
            Name := Trim (CopyLeft (S [I], J - 1), SPACE);
            Body := Trim (CopyFrom (S [I], J + 1), SPACE);
          end else
          if (Name <> '') and (S [I] <> '') and (S [I][1] in SPACE) then // header-line continuation
            Body := Body + S [I] else
            begin
              if Name <> '' then
                AddField (Name, Body);
              Name := '';
            end;
      end;
    if Name <> '' then
      AddField (Name, Body);
  End;


{                                                                              }
{ Miscellaneous header fields                                                  }
{                                                                              }

{ TDateField                                                                   }
Constructor TDateField.Create (const D : TDateTime);
  Begin
    inherited Create;
    GMTDateTime := LocalTimeToGMTTime (D);
  End;

Function TDateField.GetFieldBody : String;
  Begin
    Result := GMTDateTimeToRFC1123DateTime (GMTDateTime);
  End;

Procedure TDateField.SetFieldBody (const FieldBody : String);
  Begin
    GMTDateTime := RFCDateTimeToGMTDateTime (FieldBody);
  End;

Function TDateField.GetFieldName : String;
  Begin
    Result := 'Date';
  End;

{ TEMailField                                                                  }
Function TEMailField.GetFieldBody : String;
  Begin
    Result := Address + Cond (Name <> '', ' (' + Name + ')', '');
  End;

{ Allowed Usenet News formats (RFC850):                                        }
{     mark@cbosgd.UUCP                                                         }
{     mark@cbosgd.UUCP (Mark Horton)                                           }
{     Mark Horton <mark@cbosgd.UUCP>                                           }
{ Additional formats supported:                                                }
{     <david@e.co.za> David Butler                                             }
{     "David Butler" david@e.co.za                                             }
{     david@e.co.za "David Butler"                                             }
Procedure DecomposeEMailField (const F : String; var EMailAddress, Name : String);
  Begin
    Name := CopyBetween (F, '"', '"');
    if Name = '' then
      begin
        Name := CopyBetween (F, '(', ')');
        if Name = '' then
          begin
            Name := Trim (CopyBefore (F, '<', False));
            if Name = '' then
              Name := Trim (CopyAfter (F, '>'));
          end;
      end;

    EmailAddress := CopyBetween (F, '<', '>');
    if EmailAddress = '' then
      begin
        EmailAddress := Trim (CopyBefore (F, [',', ':', ';', ' ', '"', '('], False));
        if EmailAddress = '' then
          begin
            EmailAddress := Trim (CopyAfter (F, '"'));
            if EmailAddress = '' then
              EmailAddress := F;
          end;
      end;
  End;

Procedure TEMailField.SetFieldBody (const FieldBody : String);
  Begin
    DecomposeEMailField (FieldBody, Address, Name);
  End;

Function UUHeaderLine (const S : String; var FileName : String) : Boolean;
var T : String;
    R : Boolean;
  Begin
    T := LowerCase (S);
    Result := (Copy (T, 1, 6) = 'begin ') or (PosNext ('--- cut here ---', T) > 0);
    if Result then
      begin
        if Copy (T, 1, 6) = 'begin ' then
          T := TrimLeft (TrimLeft (CopyFrom (T, 7), [' ']), ['0'..'9']);
        Repeat
          T := TrimLeft (T, ['-', ' ']);
          R := Copy (T, 1, 8) = 'cut here';
          if R then
            T := CopyFrom (T, 9);
        Until not R;
        FileName := T;
      end else
      FileName := '';
  End;

{                                                                              }
{ MIME Decoding                                                                }
{                                                                              }

{ Examples from RFC2045:                                                       }
{   "MIME-Version: 1.0                                                         }
{    MIME-Version: 1.0 (produced by MetaSend Vx.x)                             }
{    MIME-Version: (produced by MetaSend Vx.x) 1.0                             }
{    MIME-Version: 1.(produced by MetaSend Vx.x)0                           "  }
Procedure DecomposeMimeVersionField (const F : String; var MajorVersion, MinorVersion : Integer; var Comment : String);
var S : String;
  Begin
    Comment := CopyBetween (F, '(', ')', False);
    S := Remove ('(', ')', F);
    MajorVersion := StrToIntDefault (Trim (CopyBefore (S, '.')), 0);
    MinorVersion := StrToIntDefault (Trim (CopyAfter (S, '.')), 0);
  End;

Procedure TMimeVersionField.SetFieldBody (const FieldBody : String);
  Begin
    DecomposeMimeVersionField (FieldBody, MajorVersion, MinorVersion, Comment);
    inherited SetFieldBody (FieldBody);
  End;

Function TMimeVersionField.GetFieldBody : String;
  Begin
    Result := IntToStr (MajorVersion) + '.' + IntToStr (MinorVersion) +
              Cond (Comment <> '', ' (' + Comment + ')', '');
  End;

Function TMimeVersionField.GetFieldName : String;
  Begin
    Result := 'MIME-Version';
  End;

{ RFC2045:                                                                     }
{   " content := "Content-Type" ":" type "/" subtype                           }
{                *(";" parameter)                                           "  }
Function MimeParameterValue (const Parameters, ParameterName : String) : String;
  Begin
    Result := CopyBetween (LowerCase (Parameters), ParameterName + '=', ';', True);
    if IsQuotedString (Result, ['"']) then
      Result := Copy (Result, 2, Length (Result) - 2);
  End;

Procedure DecomposeMimeContentTypeField (const F : String; var MajorType, SubType, Parameters : String);
  Begin
    MajorType := LowerCase (CopyBefore (F, '/', True));
    SubType := LowerCase (CopyBetween (F, '/', ';', True));
    Parameters := CopyAfter (F, ';');
  End;

Procedure TMimeContentTypeField.SetFieldBody (const FieldBody : String);
  Begin
    DecomposeMimeContentTypeField (FieldBody, MajorType, SubType, Parameters);
  End;

{ From RFC2045 (Multipurpose Internet Mail Extensions (MIME) Part One: Format  }
{   of Internet Message Bodies)                                                }
{                                                                              }
{         encoding := "Content-Transfer-Encoding" ":" mechanism                }
{         mechanism := "7bit" / "8bit" / "binary" /                            }
{                      "quoted-printable" / "base64" /                         }
{                      ietf-token / x-token                                    }
{         x-token := "x-uuencode"                                              }
Procedure TMimeContentTransferEncodingField.SetFieldBody (const FieldBody : String);
var B : String;
  Begin
    B := LowerCase (FieldBody);
    if B = '7bit' then
      TransferEncoding := te_7bit else
    if B = '8bit' then
      TransferEncoding := te_8bit else
    if B = 'binary' then
      TransferEncoding := te_binary else
    if B = 'quoted-printable' then
      TransferEncoding := te_quoted_printable else
    if B = 'base64' then
      TransferEncoding := te_base64 else
    if B = 'x-uuencode' then
      TransferEncoding := te_x_uuencode else
      TransferEncoding := te_none;
  End;

Procedure TMimeMessageHeader.AddField (const FieldName, FieldBody : String);
var S : String;
  Begin
    S := LowerCase (FieldName);
    if FieldName = 'mime-version' then
      AddField (TMimeVersionField.Create (FieldBody)) else
    if FieldName = 'content-type' then
      AddField (TMimeContentTypeField.Create (FieldBody)) else
    if FieldName = 'content-transfer-encoding' then
      AddField (TMimeContentTransferEncodingField.Create (FieldBody)) else
      inherited AddField (FieldName, FieldBody);
  End;

Procedure TMimeMessageHeader.Decode (const Header : String);
var Ver : TMimeVersionField;
    CT : TMimeContentTypeField;
  Begin
    inherited Decode (Header);

    FContentDisposition := GetFieldBody ('Content-Disposition');
    Ver := TMimeVersionField (GetField ('MIME-Version'));
    CT := TMimeContentTypeField (GetField ('Content-Type'));
    FIsMimeMessage := Assigned (Ver) and (Ver.MajorVersion = 1) or Assigned (CT) or (FTransferEncoding <> te_none);

    if FIsMimeMessage then
      begin
        FBoundary := MimeParameterValue (CT.Parameters, 'boundary');
      end;
  End;



{                                                                              }
{ TMailMessageDecoder                                                          }
{                                                                              }
Constructor TMailMessageDecoder.Create;
  Begin
    inherited Create;
    FInlineUUDecode := True;
  End;

{ See RFC2045                                                                  }
Function DecodeQuotedPrintableText (const T : String) : String;
var I : Integer;
  Begin
    Result := Remove ([#0..#8, #11..#12, #14..#31], T);

    if Result <> '' then
      begin
        // Trim trailing soft line break (=)
        if Result [Length (Result)] = '=' then
          SetLength (Result, Length (Result) - 1);

        // =HH hex representation of char
        I := 0;
        Repeat
          I := PosNext ('=', Result, I);
          if (I > 0) and (I <= Length (Result) - 2)
             and (Result [I + 1] in cs_HexDigit)
             and (Result [I + 2] in cs_HexDigit) then
            begin
              Result [I] := Char (HexToInt (Result [I + 1] + Result [I + 2]));
              System.Delete (Result, I + 1, 2);
            end;
        Until I = 0;

        // Trim trailing spaces
        Result := TrimRight (Result, cs_WhiteSpace);
      end;
  End;

Function DecodeTransferEncoding (const T : String; const TransferEncoding : TTransferEncoding) : String;
  Begin
    Case TransferEncoding of
      te_quoted_printable : Result := DecodeQuotedPrintableText (T);
      te_base64           : Result := MIMEDecode (T);
      te_x_uuencode       : Result := UUDecode (T);
      else Result := T;
    end;
  End;

Procedure TMultiPartBodyPart.AddPart (const BodyPart : TMessageBodyPart);
var L : Integer;
  Begin
    L := Length (Body);
    SetLength (Body, L + 1);
    Body [L] := BodyPart;
  end;

Procedure TMailMessageDecoder.Decode (const Header : String; const Body : TStream);
var Line : String;
    TransferEncoding : TTransferEncoding;

  Function ExtractMimeHeader : TMimeMessageHeader;
  var H, S : String;
    Begin
      H := '';
      Repeat
        S := Body.ExtractLine;
        if S <> '' then
          H := H + S + CRLF;
      Until S = '';
      Result := TMimeMessageHeader.Create;
      Result.Decode (H);
    End;

  Function ExtractRest : String;
    Begin
      Result := '';
      While not Body.EOF do
        begin
          Line := Body.ExtractLine;
          Result := Result + DecodeTransferEncoding (Line, TransferEncoding);
        end;
    End;

  Function ExtractToBoundary (const Boundary : String; var OpenBoundary : Boolean) : String;
    Begin
      Result := '';
      OpenBoundary := False;
      While not Body.EOF do
        begin
          Line := Body.ExtractLine;
          OpenBoundary := Line = '--' + Boundary;
          if OpenBoundary or (Line = '--' + Boundary + '--') then
            exit;
          Result := Result + DecodeTransferEncoding (Line, TransferEncoding);
        end;
    End;

  Function ExtractMimeBodyPart (const ParentHdr : TMimeMessageHeader) : TMessageBodyPart;
  var Hdr  : TMimeMessageHeader;
      CT   : TMimeContentTypeField;
      Open : Boolean;
    Begin
      if ParentHdr.FBoundary <> '' then
        begin
          CT := TMimeContentTypeField (ParentHdr.GetField ('Content-Type'));
          if CT.MajorType = 'multipart' then
            Result := TMultipartBodyPart.Create else
            Result := TMessageBodyPart.Create;
          ExtractToBoundary (ParentHdr.FBoundary, Open);
          While Open do
            begin
              Hdr := ExtractMimeHeader;
              if CT.MajorType = 'multipart' then
                begin
                  TMultipartBodyPart (Result).AddPart (ExtractMimeBodyPart (Hdr));
                  ExtractToBoundary (ParentHdr.FBoundary, Open);
                end else
                Result.Body := Result.Body + ExtractToBoundary (ParentHdr.FBoundary, Open);
            end;
        end else
        begin
          Result := TMessageBodyPart.Create;
          Result.Body := ExtractRest;
        end;
    End;

  Function NextLine : String;
    Begin
      Result := Body.ExtractLine;
    End;

  Function ExtractInlineUUEncoding (var Name : String; var DataBad : Boolean) : String;
  var S, T : String;
      R    : Boolean;
      F    : Integer;
    Begin
      // UU header
      Repeat
        S := NextLine;
        R := UUHeaderLine (S, T);
        if R and (Name = '') then
          Name := T;
        R := R or (Length (T) <= 1);
      Until not R or Body.EOF;

      // UU data
      Result := '';
      DataBad := False;
      if not Body.EOF then
        Repeat
          if Length (T) > 1 then
            if LowerCase (Copy (T, 1, 3)) = 'end' then
              break else
              begin
                System.Delete (T, 1, 1);
                try
                  Result := Result + UUDecode (T);
                except
                  DataBad := True;
                end;
              end;
          R := Body.EOF;
          if not R then
            T := NextLine;
        Until R;

      // UU footer
      F := Body.Position;
      T := LowerCase (NextLine);
      if PosNext ('cut here', T) = 0 then
        Body.Position := F;
    End;

  Function ExtractRFC822BodyPart : TMessageBodyPart;
  var S, N    : String;
      Data    : String;
      DataBad : Boolean;
      MB      : TMultiPartBodyPart;
      B       : TMessageBodyPart;
      R       : Boolean;
    Begin
      MB := TMultiPartBodyPart.Create;
      While not Body.EOF do
        begin
          Data := '';
          R := False;
          While not Body.EOF and not R do
            begin
              S := NextLine;
              R := FInlineUUDecode and UUHeaderLine (S, N);
              if not R then
                if not Assigned (FOnTextMessageLine) then
                  Data := Data + S + CRLF else
                  if FOnTextMessageLine (S) then
                    Data := Data + S + CRLF;
            end;

          if Data <> '' then
            begin
              B := TTextBodyPart.Create;
              B.Body := Data;
              MB.AddPart (B);
            end;

          if R then
            begin
              B := TAttachmentBodyPart.Create;
              B.Body := ExtractInlineUUEncoding (N, DataBad);
              TAttachmentBodyPart (B).Name := N;
            end;
        end;
      Result := MB;
    End;

var MHdr : TMimeMessageHeader;
    TE   : TMimeContentTransferEncodingField;
  Begin
    //                                                             Decode header
    MHdr := TMimeMessageHeader.Create;
    MHdr.Decode (Header);
    FHeader := MHdr;

    //                                                               Decode body
    if MHdr.IsMimeMessage then
      begin
        TE := TMimeContentTransferEncodingField (MHdr.GetField ('Content-Transfer-Encoding'));
        if Assigned (TE) then
          TransferEncoding := TE.TransferEncoding else
          TransferEncoding := te_None;
        FBody := ExtractMimeBodyPart (MHdr);
      end else
      FBody := ExtractRFC822BodyPart;
  End;




{                                                                              }
{ HTML Encoding                                                                }
{                                                                              }

{ Text encoding                                                                }
Function htmlSafeText (const Txt : String) : String;
  Begin
    Result := Replace ('&', htmlAmp, Txt);
    Result := Replace ('<', htmlLT, Result);
    Result := Replace ('>', htmlGT, Result);
    Result := Replace ('"', htmlQuote, Result);
    Result := Remove ([#0..#8, #11..#12, #14..#31], Result);
  End;

Function htmlChar (const Value : Byte) : String;
  Begin
    Result := '&#' + IntToHex (Value) + ';';
  End;

{ Tag encoding                                                                 }
Function htmlSafeQuotedText (const Txt : String; const AlwaysQuote : Boolean = False) : String;
  Begin
    Result := htmlSafeText (Txt);
    if AlwaysQuote or (PosNext ([' ', '''', '"'], Result) > 0) then
      Result := '"' + Replace ('"', htmlQuote, Result) + '"';
  End;

Function ValueAttr (const Attr, Value, DefaultValue : String) : String;
  Begin
    if Value = DefaultValue then
      Result := '' else
      Result := Attr + '=' + htmlSafeQuotedText (Value);
  End;

Function TagAttr (const Attrs : Array of String) : String;
var I : Integer;
  Begin
    Result := '';
    For I := 0 to High (Attrs) do
      if Attrs [I] <> '' then
        Result := Result + Cond (Result <> '', ' ', '') + Attrs [I];
  End;

Function htmlElement (const S : String) : String;
  Begin
    Result := '<' + S + '>';
  End;

Function Tag (const Tag : String; const Attrs : Array of String; const Txt : String) : String;
var Attr : String;
  Begin
    Attr := TagAttr (Attrs);
    Result := htmlElement (Tag + Cond (Attr <> '', ' ' + Attr, '')) +
              Txt +
              htmlElement ('/' + Tag);
  End;

{ HTML Document                                                                }
Function htmlComment (const Comment : String) : String;
  Begin
    Result := htmlElement ('-- ' + Comment + ' --') + CRLF;
  End;

Function htmlMetaHeader (const Field, Content : String) : String;
  Begin
    if Content = '' then
      Result := '' else
      Result := htmlElement ('meta http-equiv="' + Field + '" content="' + Content + '"') + CRLF;
  End;

Function htmlHead (const Title : String; const Description : String = '';
         const Keywords : String = '';
         const AutoRefreshTime : Integer = -1; const AutoRefreshURL : String = '') : String;
  Begin
    Result := Tag ('HEAD', [], CRLF +
        Cond (Title <> '', Tag ('TITLE', [], CRLF + Title + CRLF), '') + CRLF +
        htmlMetaHeader ('content-type', 'text/html; charset=iso-8859-1') +
        htmlMetaHeader ('title', Title) +
        htmlMetaHeader ('description', Description) +
        htmlMetaHeader ('keywords', Keywords) +
        Cond (AutoRefreshTime >= 0,
            htmlMetaHeader ('refresh', IntToStr (AutoRefreshTime) + Cond (AutoRefreshURL <> '', ';URL=' + AutoRefreshURL, '')), ''));
  End;

Function htmlBody (const Txt : String; const BackColour : String = '';
         const TextColour : String = ''; const LinkColour : String = '') : String;
  Begin
    Result := Tag ('BODY', [ValueAttr ('BGCOLOR', BackColour, ''),
                            ValueAttr ('TEXT', TextColour, ''),
                            ValueAttr ('LINK', LinkColour, '')],
                   CRLF + Txt + CRLF);
  End;

Function htmlDocument (
         const Title, Txt : String;
         const Description : String = '';
         const Keywords : String = '';
         const BodyBackColour : String = '';
         const BodyTextColour : String = '';
         const BodyLinkColour : String = '') : String;
  Begin
    Result := htmlElement ('!doctype HTML public "-//W3O//DTD W3 HTML 3.2//EN"') + CRLF +
              Tag ('HTML', [], CRLF +
                htmlHead (Title, Description, Keywords) + CRLF +
                htmlBody (Txt, BodyBackColour, BodyTextColour, BodyLinkColour) + CRLF) + CRLF;
  End;

{ Text format                                                                  }
Function htmlPreformattedText (const Txt : String; const WidthChars : Integer = 0) : String;
  Begin
    Result := Tag ('PRE', [ValueAttr ('WIDTH', IntToStr (WidthChars), '0')], CRLF + Txt + CRLF);
  End;

Function htmlBold (const Txt : String) : String;
  Begin
    Result := Tag ('B', [], Txt);
  End;

Function htmlItalic (const Txt : String) : String;
  Begin
    Result := Tag ('I', [], Txt);
  End;

Function htmlUnderline (const Txt : String) : String;
  Begin
    Result := Tag ('U', [], Txt);
  End;

Function htmlStrikethrough (const Txt : String) : String;
  Begin
    Result := Tag ('STRIKE', [], Txt);
  End;

Function htmlTeletypeFont (const Txt : String) : String;
  Begin
    Result := Tag ('TT', [], Txt);
  End;

Function htmlFont (const Txt : String; const FontSize : Integer = 0; const FontColour : String = ''; const FontFace : String = '') : String;
  Begin
    Result := Tag ('FONT', [ValueAttr ('FACE', FontFace, ''),
                            ValueAttr ('SIZE', IntToStr (FontSize), '0'),
                            ValueAttr ('COLOR', FontColour, '')], Txt);
  End;

{ Formatting                                                                   }
Function htmlCenter (const Txt : String) : String;
  Begin
    Result := Tag ('CENTER', [], Txt);
  End;

{ Tables                                                                       }
Function htmlTable (const Txt : String; const Border : Boolean;
         const Width : String; const BackColour : String;
         const CellSpacing : Integer; const CellPadding : Integer) : String;
  Begin
    Result := Tag ('TABLE', [ValueAttr ('WIDTH', Width, ''),
                             ValueAttr ('BORDER', Cond (Border, '1', '0'), '0'),
                             ValueAttr ('BGCOLOR', BackColour, ''),
                             ValueAttr ('CELLSPACING', IntToStr (CellSpacing), '2'),
                             ValueAttr ('CELLPADDING', IntToStr (CellPadding), '1')],
                             CRLF + Txt + CRLF);
  End;

{ Works with both IE and Netscape                                              }
{ Need to set the colours of the rows/cols to the BackColour to make it look   }
{ the same on both.                                                            }
Function htmlBorderedTable (const Txt : String; const Width : String;
         const FrameWidth : Integer; const FrameColour : String;
         const CellPadding : Integer) : String;
  Begin
    Result := htmlTable (htmlRow (htmlCol (htmlTable (Txt, False, '100%', FrameColour, FrameWidth, CellPadding))),
                           False, Width, FrameColour, 0, 0);
  End;

Function htmlRow (const Txt : String; const BackColour : String) : String;
  Begin
    Result := Tag ('TR', [ValueAttr ('BGCOLOR', BackColour, '')], Txt) + CRLF;
  End;

Function htmlCol (const Txt : String; const ColSpan : Integer = 1;
         const Width : String = ''; const BackColour : String = '') : String;
  Begin
    Result := Tag ('TD', [ValueAttr ('BGCOLOR', BackColour, ''),
                          ValueAttr ('WIDTH', Width, ''),
                          ValueAttr ('COLSPAN', IntToStr (ColSpan), '1')], Txt);
  End;

{ Other Tags                                                                   }
Function htmlImage (const Img : String; const Align : String = '';
         const AltText : String = '';
         const Width : Integer = 0; const Height : Integer = 0;
         const Border : Boolean = True) : String;
  Begin
    Result := htmlElement ('IMG ' + TagAttr ([
        ValueAttr ('SRC', Img, ''),
        ValueAttr ('ALIGN', Align, ''),
        ValueAttr ('ALT', AltText, ''),
        ValueAttr ('BORDER', Cond (Border, '1', '0'), '1'),
        ValueAttr ('WIDTH', IntToStr (Width), '0'),
        ValueAttr ('HEIGHT', IntToStr (Height), '0')])) + CRLF;
  End;

Function htmlLink (const Link : String; const Txt : String) : String;
  Begin
    Result := Tag ('A', [ValueAttr ('HREF', Link, '')], Txt);
  End;



{                                                                              }
{ HTML Representations                                                         }
{                                                                              }

Constructor AhtmlRenderer.Create (const Parent : AhtmlRenderer);
  Begin
    inherited Create;
    if Assigned (Parent) then
      begin
        FontSize := Parent.FontSize;
        FontColour := Parent.FontColour;
        FontBackColour := Parent.FontBackColour;
        Attributes := Parent.Attributes;
      end else
      begin
        FontSize := 3;
      end;
  End;

{ ThtmlComponent                                                               }
Procedure ThtmlComponent.Render (const Renderer : AhtmlRenderer);
  Begin
  End;

{ ThtmlText                                                                    }
Constructor ThtmlText.Create (const Text : String);
  Begin
    inherited Create;
    FText := Text;
  End;

Function ThtmlText.GetAsHTML : String;
  Begin
    Result := FText;
  End;

{ ThtmlEntity                                                                  }
Constructor ThtmlEntity.Create (const Entity : String);
  Begin
    inherited Create;
    FEntity := Entity;
  End;

Function ThtmlEntity.GetAsHTML : String;
  Begin
    Result := '&' + FEntity;
  End;

{ ThtmlConcatenation                                                           }
Constructor ThtmlConcatenation.Create (const Left, Right : AhtmlComponent);
  Begin
    inherited Create;
    SetLength (FComponents, 2);
    FComponents [0] := Left;
    FComponents [1] := Right;
  End;

Destructor ThtmlConcatenation.Destroy;
  Begin
    FreeObjectArray (FComponents);
    inherited Destroy;
  End;

Procedure ThtmlConcatenation.Add (const C : AhtmlComponent);
  Begin
    Append (ObjectArray (FComponents), C);
  End;

Function ThtmlConcatenation.Count : Integer;
  Begin
    Result := Length (FComponents);
  End;

Function ThtmlConcatenation.GetComponent (const Idx : Integer) : AhtmlComponent;
  Begin
    Result := FComponents [Idx];
  End;

Function ThtmlConcatenation.GetAsHTML : String;
var I : Integer;
  Begin
    Result := '';
    For I := 0 to Length (FComponents) - 1 do
      Result := Result + FComponents [I].GetAsHTML;
  End;

Procedure ThtmlConcatenation.Render (const Renderer : AhtmlRenderer);
var I : Integer;
  Begin
    For I := 0 to Length (FComponents) - 1 do
      FComponents [I].Render (Renderer);
  End;

{ ThtmlAttributes                                                              }
Procedure ThtmlAttributes.AddAttribute (const Attr, Value : String);
var L : Integer;
  Begin
    L := Length (FAttr);
    SetLength (FAttr, L + 1);
    FAttr [L].Attr := Attr;
    FAttr [L].Value := Value;
  End;

Function ThtmlAttributes.GetAsHTML : String;
var I : Integer;
  Begin
    Result := '';
    For I := 0 to Length (FAttr) - 1 do
      Result := Result + Cond (Result <> '', ' ', '') +
                FAttr [I].Attr + '=' +
                htmlSafeQuotedText (FAttr [I].Value);
  End;

Function ThtmlAttributes.IndexOf (const Attr : String) : Integer;
var I : Integer;
  Begin
    For I := 0 to Length (FAttr) - 1 do
      if FAttr [I].Attr = Attr then
        begin
          Result := I;
          exit;
        end;
    Result := -1;
  End;

Function ThtmlAttributes.AsString (const Attr : String; const DefaultValue : String) : String;
var I : Integer;
  Begin
    I := IndexOf (Attr);
    if I = -1 then
      Result := DefaultValue else
      Result := FAttr [I].Value;
  End;

Function ThtmlAttributes.AsInteger (const Attr : String; const DefaultValue : Integer) : Integer;
var I : Integer;
  Begin
    I := IndexOf (Attr);
    if I = -1 then
      Result := DefaultValue else
      try
        Result := StrToInt (FAttr [I].Value);
      except
        Result := DefaultValue;
      end;
  End;

Function ThtmlAttributes.Duplicate : ThtmlAttributes;
  Begin
    Result := ThtmlAttributes.Create;
    Result.FAttr := FAttr;
  End;

{ ThtmlElement                                                                 }
Constructor ThtmlElement.Create (const Name : String; const Attributes : ThtmlAttributes);
  Begin
    inherited Create;
    FName := Name;
    FAttr := Attributes;
  End;

Function ThtmlElement.GetAsHTML : String;
var S : String;
  Begin
    if Assigned (FAttr) then
      S := FAttr.GetAsHTML else
      S := '';
    Result := '<' + FName + Cond (S <> '', ' ' + S, '') + '>';
  End;

Destructor ThtmlElement.Destroy;
  Begin
    FreeAndNil (FAttr);
    inherited Destroy;
  End;

Function ThtmlElement.GetName : String;
  Begin
    if not CloseElement then
      Result := FName else
      Result := CopyFrom (FName, 2);
  End;

Function ThtmlElement.GetCloseElement : Boolean;
  Begin
    Result := (FName <> '') and (FName [1] = '/');
  End;

Function ThtmlElement.AttrAsString (const Attr : String; const DefaultValue : String) : String;
  Begin
    if Assigned (FAttr) then
      Result := FAttr.AsString (Attr, DefaultValue) else
      Result := DefaultValue;
  End;

Function ThtmlElement.AttrAsInteger (const Attr : String; const DefaultValue : Integer) : Integer;
  Begin
    if Assigned (FAttr) then
      Result := FAttr.AsInteger (Attr, DefaultValue) else
      Result := DefaultValue;
  End;

Procedure ThtmlElement.Render (const Renderer : AhtmlRenderer);
  Begin
//    raise EAbstractError.Create ('Can not render for ' + FName);
  End;

{ ThtmlTag                                                                     }
Constructor ThtmlTag.Create (const OpenElement, CloseElement : AhtmlElement; const Enclosed : AhtmlComponent);
  Begin
    inherited Create;
    FOpenElement := OpenElement;
    FEnclosed := Enclosed;
    FCloseElement := CloseElement;
  End;

Function ThtmlTag.GetAsHTML : String;
  Begin
    Result := FOpenElement.GetAsHTML;
    if Assigned (FEnclosed) then
      Result := Result + FEnclosed.GetAsHTML;
    if Assigned (FCloseElement) then
      Result := Result + FCloseElement.GetAsHTML;
  End;

Function ThtmlTag.GetEnclosed : AhtmlComponent;
  Begin
    Result := FEnclosed;
  End;

Procedure ThtmlTag.Render (const Renderer : AhtmlRenderer);
  Begin
    if Assigned (FEnclosed) then
      FEnclosed.Render (Renderer);
  End;

Destructor ThtmlTag.Destroy;
  Begin
    FreeAndNil (FCloseElement);
    FreeAndNil (FEnclosed);
    FreeAndNil (FOpenElement);
    inherited Destroy;
  End;

{ ThtmlUnknownElement                                                          }
Constructor ThtmlUnknownElement.Create (const Element : String);
  Begin
    inherited Create;
    FElement := Element;
  End;

Function ThtmlUnknownElement.GetAsHTML : String;
  Begin
    Result := '<' + FElement + '>';
  End;

{ ThtmlASPCode                                                                 }
Constructor ThtmlASPCode.Create (const Code : String);
  Begin
    inherited Create;
    FCode := Code;
  End;

Function ThtmlASPCode.GetAsHTML : String;
  Begin
    Result := '<%' + FCode + '%>';
  End;

{ ThtmlComment                                                                 }
Constructor ThtmlComment.Create (const Comment : String);
  Begin
    inherited Create;
    FComment := Comment;
  End;

Function ThtmlComment.GetAsHTML : String;
  Begin
    Result := '<-- ' + FComment + ' -->';
  End;

{ ThtmlDTDTag                                                                  }
Constructor ThtmlDTDTag.Create (const Text : String);
  Begin
    inherited Create;
    FText := Text;
  End;

Function ThtmlDTDTag.GetAsHTML : String;
  Begin
    Result := '<!' + FText + '>';
  End;

{ ThtmlXMLTag                                                                  }
Constructor ThtmlProcessingInstructions.Create (const PI : String);
  Begin
    inherited Create;
    FPI := PI;
  End;

Function ThtmlProcessingInstructions.GetAsHTML : String;
  Begin
    Result := '<?' + FPI + '>';
  End;

{ ThtmlSSITag                                                                  }
Constructor ThtmlSSITag.Create (const Text : String);
  Begin
    inherited Create;
    FText := Text;
  End;

Function ThtmlSSITag.GetAsHTML : String;
  Begin
    Result := '<!--#' + FText + '-->';
  End;



{                                                                              }
{ HTML Decoding                                                                }
{                                                                              }
type
  TColEntry = record
    htmlColour : String;
    Colour     : TColor;
  end;

const
  htmlPredefinedCols = 18;
  htmlColTable : Array [1..htmlPredefinedCols] of TColEntry = (
      { HTML4.0 Colours }
      (htmlColour: 'BLACK';   Colour:clBlack),  (htmlColour: 'GREEN';  Colour:clGreen),
      (htmlColour: 'SILVER';  Colour:clSilver), (htmlColour: 'LIME';   Colour:clLime),
      (htmlColour: 'GRAY';    Colour:clGray),   (htmlColour: 'OLVE';   Colour:clOlive),
      (htmlColour: 'WHITE';   Colour:clWhite),  (htmlColour: 'YELLOW'; Colour:clYellow),
      (htmlColour: 'MAROON';  Colour:clMaroon), (htmlColour: 'NAVY';   Colour:clNavy),
      (htmlColour: 'RED';     Colour:clRed),    (htmlColour: 'BLUE';   Colour:clBlue),
      (htmlColour: 'PURPLE';  Colour:clPurple), (htmlColour: 'TEAL';   Colour:clTeal),
      (htmlColour: 'FUCHSIA'; Colour:clFuchsia),(htmlColour: 'AQUA';   Colour:clAqua),
      { Custom extentions }
      (htmlColour: 'LIGHTGRAY'; Colour:clLtGray),  (htmlColour: 'DARKGRAY'; Colour:clDkGray));

Function htmlColour (const Colour : String) : TColor;
var I : Integer;
    C : Integer;
  Begin
    if Colour [1] = '#' then
      begin
        C := HexToInt (CopyFrom (Colour, 2));
        Result := ((C and $FF) shl 16) or
                  (C and $FF00) or
                  ((C and $FF0000) shr 16);
      end else
      begin
        For I := 1 to htmlPredefinedCols do
          begin
            if htmlColTable [I].htmlColour = Colour then
              begin
                Result := htmlColTable [I].Colour;
                exit;
              end;
          end;
        raise EhtmlParser.Create ('Invalid HTML colour.');
      end;
  End;

Procedure htmlDecodeSize (const S : String; var Size : Integer; var SizeType : ThtmlSizeType);
  Begin
    if S <> '' then
      try
        Case S [Length (S)] of
          '%' :
            begin
              SizeType := stPercent;
              Size := StrToInt (CopyLeft (S, Length (S) - 1));
            end;
          '*' :
            begin
              SizeType := stRelative;
              Size := StrToInt (CopyLeft (S, Length (S) - 1));
            end;
          else
            begin
              SizeType := stPixels;
              Size := StrToInt (S);
            end;
        end;
      except
        SizeType := stNone;
        Size := 0;
      end else
      begin
        SizeType := stNone;
        Size := 0;
      end;
  End;

{ Translate                                                                    }
Function Translate (const Ch : Char; const TranslationTable : String) : Char;
  Begin
    Assert (Length (TranslationTable) = 256, 'Translation table must be 256 characters long.');
    Result := TranslationTable [Ord (Ch) + 1];
  End;

Function Translate (const S : String; const TranslationTable : String) : String;
var I : Integer;
  Begin
    Assert (Length (TranslationTable) = 256, 'Translation table must be 256 characters long.');
    Result := S;
    For I := 1 to Length (Result) do
      Result [I] := TranslationTable [Ord (Result [I]) + 1];
  End;



{                                                                              }
{ ThtmlParser                                                                  }
{   See http://www.w3.org/TR/1999/REC-html401-19991224/index/elements.html     }
{                                                                              }
Function ThtmlParser.InterpretedTag (const Element : AhtmlElement) : AhtmlElement;
  Begin
    Result := Element;
  End;

Function ThtmlParser.InterpretedTag (const OpenElement, CloseElement : AhtmlElement; const Enclosed : AhtmlComponent) : AhtmlTag;
  Begin
    Result := ThtmlTag.Create (OpenElement, CloseElement, Enclosed);
  End;

Function ThtmlParser.InterpretedText (const Text : String) : AhtmlComponent;
  Begin
    Result := ThtmlText.Create (Text);
  End;

Function ThtmlParser.InterpretedEntity (const Entity : String) : AhtmlComponent;
  Begin
    Result := ThtmlEntity.Create (Entity);
  End;

{                                                                              }
{ ThtmlParser.GetNextToken                                                     }
{   Lexical tokeniser                                                          }
{                                                                              }
Function ThtmlParser.GetNextToken : AhtmlComponent;

  { "<%"<asp code>"%>"                                                         }
  Function ExtractASP : AhtmlComponent;
    Begin
      Result := ThtmlASPCode.Create (FS.ExtractDelim ('%>', True));
    End;

  { "<!-- "<comment>" -->"                                                     }
  Function ExtractComment : AhtmlComponent;
    Begin
      Result := ThtmlComment.Create (FS.ExtractDelim ('-->', True));
    End;

  { "<!"<dtd tag>">"                                                           }
  Function ExtractDTD : AhtmlComponent;
    Begin
      Result := ThtmlDTDTag.Create (FS.ExtractDelim ('>', True));
    End;

  { "<?"<processing instructions>">"                                           }
  { "<?"<xml tag>"?>"                                                          }
  Function ExtractProcessingInstructions : AhtmlComponent;
    Begin
      Result := ThtmlProcessingInstructions.Create (FS.ExtractDelim ('>', True));
    End;

  { "<!--#"<server side includes>"-->"                                         }
  Function ExtractSSI : AhtmlComponent;
    Begin
      Result := ThtmlSSITag.Create (FS.ExtractDelim ('-->', True));
    End;

  // Pre:  S <> ''
  // Post: Returns '' if entity not recognised.
  Function DecodeEntity (const S : String) : String;
  var U : String;
    Begin
      if S [1] = '#' then
        try
          Result := Char (StrToInt (Replace ('x', '$', CopyFrom (S, 2))));
        except
          Result := '';
        end else
        begin
          U := LowerCase (S);
          if S = 'amp' then
            Result := '&' else
          if S = 'lt' then
            Result := '<' else
          if S = 'gt' then
            Result := '>' else
          if S = 'quote' then
            Result := '"' else
            Result := '';
        end;
    End;

  const
    cs_Entity = cs_Alpha;

  { "&"<entity>[";"]                                                           }
  { "&#"<entity code>[";"]                                                     }
  Function ExtractEntity : AhtmlComponent;
  var T, R : String;
    Begin
      T := FS.Extract (['#']);
      if T <> '' then
        T := T + FS.ExtractAll (cs_Numeric) else
        T := FS.ExtractAll (cs_Alpha);
      if T <> '' then
        begin
          FS.Extract ([';']);
          R := DecodeEntity (T);
          if R <> '' then
            begin
              Result := InterpretedText (R);
              exit;
            end;
          Result := InterpretedEntity (T);
        end else
        Result := InterpretedText ('&');
    End;

  Function DecodeAttrValue (const S : String) : String;
  var I    : Integer;
      T, R : String;
      U    : Boolean;
    Begin
      I := 0;
      Result := S;
      Repeat
        I := PosNext ('&', Result, I);
        if I > 0 then
          begin
            T := CopyBefore (Result, cs_Everything - cs_AlphaNumeric - ['#'], False, I + 1);
            if Length (T) >= 2 then
              begin
                R := DecodeEntity (T);
                if R <> '' then
                  begin
                    U := (Length (Result) >= I + 1 + Length (T)) and
                         (Result [I + 1 + Length (T)] = ';');
                    Delete (Result, I, Length (T) + 1 + Cond (U, 1, 0));
                    Insert (R, Result, I);
                  end;
              end;
          end;
      Until I = 0;
    End;

  { "<"<name>[SP*<attrname>["="(<attrvalue>|""""<attrvalue>"""")]]*">"         }
  Function ExtractElement : AhtmlComponent; pascal;
  const
    cs_ElementNameStart = cs_Alpha;
    cs_ElementName      = cs_AlphaNumeric + ['-', '.'];
    cs_AttrNameStart    = cs_Alpha;
    cs_AttrName         = cs_AlphaNumeric + ['-', '.'];
  var Pre,
      Name   : String;
      AName,
      AVal   : String;
      Attr   : ThtmlAttributes;
    Begin
      Pre := FS.Extract (['/']);
      Name := FS.Extract (cs_ElementNameStart);
      if Name = '' then
        begin
          Result := InterpretedText ('<' + Pre + Name);
          exit;
        end;
      Name := Pre + UpperCase (Name + FS.ExtractAll (cs_ElementName));
      FS.SkipWhiteSpace;
      Attr := nil;
      While not FS.EOF and not FS.Match ('>') do
        begin
          AName := UpperCase (FS.ExtractDelim (cs_WhiteSpace + ['>', '='], False));
          FS.SkipWhiteSpace;
          if FS.Extract ('=') <> '' then
            begin
              FS.SkipWhiteSpace;
              Pre := FS.Extract (['"', '''']);
              if Pre <> '' then
                begin
                  AVal := FS.ExtractDelim ([Pre [1]] + ['>'], False);
                  if FS.Match (Pre [1]) then
                    FS.Skip;
                end else
                AVal := FS.ExtractDelim (cs_WhiteSpace + ['>'], False);
              FS.SkipWhiteSpace;
            end else
            begin
              AVal := AName;
              AName := '';
            end;
          if not Assigned (Attr) then
            Attr := ThtmlAttributes.Create;
          Attr.AddAttribute (AName, DecodeAttrValue (AVal));
        end;
      FS.Extract;
      Result := ThtmlElement.Create (Name, Attr);
    End;

var Ch : Char;
    T  : String;
  Begin
    if not FS.EOF then
      begin
        T := FS.ExtractDelim (['<', '&']);
        if T <> '' then
          Result := InterpretedText (T) else
          begin
            Ch := FS.Extract;
            if Ch = '&' then
              Result := ExtractEntity else
            if Ch = '<' then
              begin
                if FS.Extract ('?') <> '' then
                  Result := ExtractProcessingInstructions else
                if FS.Extract ('%') <> '' then
                  Result := ExtractASP else
                if FS.Match ('!') then
                  if FS.Extract ('!--#') <> '' then
                    Result := ExtractSSI else
                    if FS.Extract ('!--') <> '' then
                      Result := ExtractComment else
                      begin
                        FS.Extract;
                        Result := ExtractDTD;
                      end
                else
                  Result := ExtractElement;
              end;
          end;
      end else
      Result := nil;

    FToken := Result;
  End;

// Concatenate component C to component R
Procedure AppendComponent (var R : AhtmlComponent; const C : AhtmlComponent);
  Begin
    if Assigned (C) then
      if not Assigned (R) then
        R := C else
        if not (R is ThtmlConcatenation) then
          R := ThtmlConcatenation.Create (R, C) else
          ThtmlConcatenation (R).Add (C);
  End;

{                                                                              }
{ ThtmlParser.ParseHTMLTree                                                    }
{   Generate a recursive tree structure from HTML.                             }
{                                                                              }
Function ThtmlParser.ParseHTMLTree (const S : TStream) : AhtmlComponent;

  { Helper functions                                                           }
  Function GetElement : AhtmlElement;
    Begin
      Result := AhtmlElement (FToken);
      GetNextToken;
    End;

  Function IsOpenElement (const Name : String = '') : Boolean;
    Begin
      Result := (FToken is AhtmlElement) and not AhtmlElement (FToken).CloseElement and
                ((Name = '') or (AhtmlElement (FToken).Name = Name));
    End;

  Function IsCloseElement (const Name : String = '') : Boolean;
    Begin
      Result := (FToken is AhtmlElement) and AhtmlElement (FToken).CloseElement and
                ((Name = '') or (AhtmlElement (FToken).Name = Name));
    End;

  Function IfCloseElement (const Name : String = '') : AhtmlElement;
    Begin
      if IsCloseElement (Name) then
        Result := GetElement else
        Result := nil;
    End;

  Function IsElement (const Name : String = '') : Boolean;
    Begin
      Result := (FToken is AhtmlElement) and
                ((Name = '') or (AhtmlElement (FToken).Name = Name));
    End;

  Function IfElementGetElement (const Name : String = '') : AhtmlElement;
    Begin
      if IsElement (Name) then
        Result := GetElement else
        Result := ThtmlElement.Create (Name);
    End;

  type
    TExitOpenTags = Set of (ot_TD, ot_TR, ot_LI, ot_P, ot_DT, ot_DD, ot_BODY);

  // Level1
  Function ExtractComponent (const ExitOpenTags : TExitOpenTags = []) : AhtmlComponent; forward;

  { Level3 Head                                                                }
  Function ExtractHead (const ExitOpenTags : TExitOpenTags) : AhtmlComponent;
  var EOpen  : AhtmlElement;
      EClose : AhtmlElement;
    Begin
      EOpen := GetElement;
      Result := nil;
      While Assigned (FToken) and not (IsCloseElement or IsOpenElement ('BODY')) do
        AppendComponent (Result, ExtractComponent (ExitOpenTags + [ot_BODY]));
      EClose := IfCloseElement ('HEAD');
      Result := InterpretedTag (EOpen, EClose, Result);
    End;

  { Level3 Paragraph                                                           }
  Function ExtractParagraph (const ExitOpenTags : TExitOpenTags) : AhtmlComponent;
  var EOpen  : AhtmlElement;
      EClose : AhtmlElement;
    Begin
      EOpen := GetElement;
      Result := nil;
      While Assigned (FToken) and not (IsCloseElement or IsElement ('P')) do
        AppendComponent (Result, ExtractComponent (ExitOpenTags + [ot_P]));
      EClose := IfCloseElement ('P');
      Result := InterpretedTag (EOpen, EClose, Result);
    End;

  { Level3 Definition lists                                                    }
  Function ExtractDefinitionList (const ExitOpenTags : TExitOpenTags) : AhtmlComponent;

    Function ExtractDefinitionDescription : AhtmlComponent;
    var EOpen  : AhtmlElement;
        EClose : AhtmlElement;
      Begin
        EOpen := GetElement;
        Result := ExtractComponent (ExitOpenTags + [ot_DT, ot_DD]);
        EClose := IfCloseElement ('DD');
        Result := InterpretedTag (EOpen, EClose, Result);
      End;

    Function ExtractDefinitionType : AhtmlComponent;
    var EOpen  : AhtmlElement;
        EClose : AhtmlElement;
      Begin
        EOpen := GetElement;
        Result := ExtractComponent (ExitOpenTags + [ot_DT, ot_DD]);
        EClose := IfCloseElement ('DT');
        Result := InterpretedTag (EOpen, EClose, Result);
      End;

  var EOpen  : AhtmlElement;
      EClose : AhtmlElement;
    Begin
      EOpen := IfElementGetElement ('DL');
      Result := nil;
      While Assigned (FToken) and not IsCloseElement do
        if IsOpenElement ('DT') then
          AppendComponent (Result, ExtractDefinitionType) else
        if IsOpenElement ('DD') then
          AppendComponent (Result, ExtractDefinitionDescription) else
          AppendComponent (Result, ExtractComponent (ExitOpenTags + [ot_DT, ot_DD]));
      EClose := IfCloseElement ('DL');
      Result := InterpretedTag (EOpen, EClose, Result);
    End;

  { Level3 Lists                                                               }
  Function ExtractList (const ExitOpenTags : TExitOpenTags) : AhtmlComponent;

    Function ExtractListItem : AhtmlComponent;
    var EOpen  : AhtmlElement;
        EClose : AhtmlElement;
      Begin
        EOpen := GetElement;
        Result := ExtractComponent (ExitOpenTags + [ot_LI]);
        EClose := IfCloseElement ('LI');
        Result := InterpretedTag (EOpen, EClose, Result);
      End;

  var EOpen  : AhtmlElement;
      EClose : AhtmlElement;
    Begin
      if IsElement ('LI') then
        EOpen := ThtmlElement.Create ('UL') else
        EOpen := GetElement;

      Result := nil;
      While Assigned (FToken) and not IsCloseElement do
        if IsOpenElement ('LI') then
          AppendComponent (Result, ExtractListItem) else
          AppendComponent (Result, ExtractComponent (ExitOpenTags + [ot_LI]));

      if Assigned (EOpen) then
        EClose := IfCloseElement (EOpen.Name) else
        EClose := nil;

      Result := InterpretedTag (EOpen, EClose, Result);
    End;

  { Level3 Tables                                                              }
  Function ExtractTable (const ExitOpenTags : TExitOpenTags) : AhtmlComponent;

    Function ExtractTableCol : AhtmlComponent;
    var EOpen  : AhtmlElement;
        EClose : AhtmlElement;
      Begin
        EOpen := GetElement;
        Result := ExtractComponent (ExitOpenTags + [ot_TD, ot_TR]);
        EClose := IfCloseElement ('TD');
        Result := InterpretedTag (EOpen, EClose, Result);
      End;

    Function ExtractTableRow : AhtmlComponent;
    var EOpen  : AhtmlElement;
        EClose : AhtmlElement;
      Begin
        EOpen := IfElementGetElement ('TR');
        Result := nil;
        While Assigned (FToken) and not (IsElement ('TABLE') or IsElement ('TR')) do
          if IsOpenElement ('TD') then
            AppendComponent (Result, ExtractTableCol) else
            AppendComponent (Result, ExtractComponent (ExitOpenTags + [ot_TD, ot_TR]));
        EClose := IfCloseElement ('TR');
        Result := InterpretedTag (EOpen, EClose, Result);
      End;

  var EOpen  : AhtmlElement;
      EClose : AhtmlElement;
    Begin
      EOpen := IfElementGetElement ('TABLE');
      Result := nil;
      While Assigned (FToken) and not IsCloseElement ('TABLE') do
        if IsOpenElement ('TR') or IsOpenElement ('TD') then
          AppendComponent (Result, ExtractTableRow) else
          AppendComponent (Result, ExtractComponent (ExitOpenTags + [ot_TD, ot_TR]));
      EClose := IfCloseElement ('TABLE');
      Result := InterpretedTag (EOpen, EClose, Result);
    End;

  { Level2 Tags                                                                }
  Function ExtractTag (const ExitOpenTags : TExitOpenTags) : AhtmlComponent;
  var EOpen  : AhtmlElement;
      EClose : AhtmlElement;
    Begin
      EOpen := GetElement;
      Result := nil;
      if Assigned (FToken) then
        begin
          Result := ExtractComponent (ExitOpenTags);
          EClose := IfCloseElement (EOpen.Name);
        end else
        EClose := nil;

      Result := InterpretedTag (EOpen, EClose, Result);
    End;

  { Level1                                                                     }
  {   Extracts and concatenates components until any closing tag or an opening }
  {   tag thats in ExitOpenTags is encountered.                                }
  Function ExtractComponent (const ExitOpenTags : TExitOpenTags = []) : AhtmlComponent;

    { Tags for which no closing element is allowed.                            }
    Function IsNonClosingTag (const TagName : String) : Boolean;
    const htmlNonClosingTags = 13;
          htmlNonClosingTag : Array [1..htmlNonClosingTags] of String = (
              'AREA', 'BASE', 'BASEFONT', 'BR', 'COL', 'FRAME', 'HR',
              'IMG', 'INPUT', 'ISINDEX', 'LINK', 'META', 'PARAM');
    var I : Integer;
      Begin
        For I := 1 to htmlNonClosingTags do
          if htmlNonClosingTag [I] = TagName then
            begin
              Result := True;
              exit;
            end;
        Result := False;
      End;

  var R : Boolean;
    Begin
      Result := nil;
      R := False;
      While not R and Assigned (FToken) do
        if FToken is AhtmlElement then
          With AhtmlElement (FToken) do
            if CloseElement then
              R := True else
              // Tags with no closing tag
              if IsNonClosingTag (Name) then
                AppendComponent (Result, InterpretedTag (GetElement)) else
              // Tags with optional close elements
              if ((Name = 'TR') and (ot_TR in ExitOpenTags)) or
                 ((Name = 'TD') and (ot_TD in ExitOpenTags)) or
                 ((Name = 'LI') and (ot_LI in ExitOpenTags)) or
                 ((Name = 'DD') and (ot_DD in ExitOpenTags)) or
                 ((Name = 'DT') and (ot_DT in ExitOpenTags)) or
                 ((Name = 'BODY') and (ot_BODY in ExitOpenTags)) or
                 ((Name = 'P') and (ot_P in ExitOpenTags)) then
                R := True else
              // Custom tag parsing
              if (Name = 'TABLE') or (Name = 'TR') or (Name = 'TD') then
                AppendComponent (Result, ExtractTable (ExitOpenTags)) else
              if (Name = 'OL') or (Name = 'UL') or (Name = 'DIR') or (Name = 'LI') then
                AppendComponent (Result, ExtractList (ExitOpenTags)) else
              if (Name = 'DL') or (Name = 'DD') or (Name = 'DT') then
                AppendComponent (Result, ExtractDefinitionList (ExitOpenTags)) else
              if Name = 'P' then
                AppendComponent (Result, ExtractParagraph (ExitOpenTags)) else
              if Name = 'HEAD' then
                AppendComponent (Result, ExtractHead (ExitOpenTags)) else
              // Normal tag parsing
                AppendComponent (Result, ExtractTag (ExitOpenTags))
        else
        // Non-Elements
          begin
            AppendComponent (Result, FToken);
            GetNextToken;
          end;
    End;

  Begin
    Result := nil;
    FS := S;
    GetNextToken;
    While Assigned (FToken) do
      begin
        AppendComponent (Result, ExtractComponent);
        if Assigned (FToken) then // invalid token
          begin
            AppendComponent (Result, FToken);
            GetNextToken;
          end;
      end;
  End;

{                                                                              }
Function ThtmlParser.ParseHTMLFlat (const S : TStream) : AhtmlComponent;
  Begin
    Result := nil;
    FS := S;
    GetNextToken;
    While Assigned (FToken) do
      begin
        AppendComponent (Result, FToken);
        GetNextToken;
      end;
  End;

end.

