{$INCLUDE cHeader.inc}
unit cStream;

interface

uses
  SysUtils,             // Exception, fmxxx constants
  Classes,              // TStream, TFileStream
  scktcomp,             // TWinSocketStream

  // Delphi Fundamentals (L0)
  cUtil;



{                                                                              }
{ Streams unit v0.25 (L0)                                                      }
{                                                                              }
{    Stream abstraction and implementations.                                   }
{                                                                              }
{ This unit is copyrighted  1999-2000 by David Butler (david@e.co.za)         }
{                                                                              }
{ This unit is part of Delphi Fundamentals, it's original name is cStreams.    }
{                                                                              }
{ I invite you to use this unit, free of charge.                               }
{ I invite you to distibute this unit, but it must be for free.                }
{ I also invite you to contribute to its development, but do not distribute    }
{ a modified copy of this file, send suggestions and bug reports to            }
{ david@e.co.za                                                                }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   01/03/99  v0.01   Initial version of TextParser (Skip, Extract and Match   }
{                     methods)                                                 }
{   11/05/99  v0.02   Whitespace functions                                     }
{   20/06/99  v0.03   Added MatchString                                        }
{                     Added Line, Col, LineDelim properties                    }
{                     199 lines interface, 207 lines implementation            }
{   26/06/99  v0.10   Moved AStream from cDataStruct unit.                     }
{                     Added ATBinaryStream. Added ReadPacked functions.        }
{                     Added ATTextStream from TTextParser in unit cTextParser. }
{                     Added Peek, Skip, OctetsRemaing and OctetsAvailable to   }
{                     AStream.                                                 }
{                     Added ExtractPacked functions                            }
{                     Moved from cTextParser and renamed TTextParser to        }
{                     ATTextStream. It now inherits from AStream to allow all  }
{                     derived streams to use it.                               }
{   03/07/99  v0.11   Added MetaCharacter support (machine dependant charsets),}
{                     WhiteSpace functions are now implemented using these.    }
{                     Added MetaString support.                                }
{   17/07/99  v0.12   Added Col property.                                      }
{   03/11/99  v0.13   Revised AStream documentation.                           }
{   04/11/99  v0.14   Overloaded methods in ATBinaryStream. Added Int64.       }
{   05/11/99  v0.15   Removed MetaCharacter support.                           }
{   06/11/99  v0.16   Revisited whole unit.                                    }
{                     Overloaded methods.                                      }
{                     Added general functions to ATStream.                     }
{                     Renamed TStringStream to TMemoryStream.                  }
{                     433 lines interface, 933 lines implementation.           }
{   09/11/99  v0.17   Moved WritePacked's implementations to cStrings unit as  }
{                     functions called Pack.                                   }
{   26/12/99  v0.18   Bug fixes. Added SelfTest procedure.                     }
{   03/01/00  v0.19   Removed FLineDelim. A line delimiter is now defined as   }
{                     EOF or either or both of CR and LF.                      }
{   08/02/00  v0.20   Added ATRecordStream.                                    }
{   27/02/00  v0.21   AStream now derives from Delphi's TStream.               }
{                     Revisited whole unit.                                    }
{                     521 lines interface, 1468 lines implementation.          }
{   05/03/00  v0.22   Renamed TMemoryStream to TStringStream.                  }
{   12/03/00  v0.23   Added TEOLType to ATTextStream.                          }
{   23/04/00  v0.24   Added TSocketStream for use in cHTTPServer.              }
{   08/05/00  v0.25   Cleaned up unit. Merged ATStream, ATBinaryStream,        }
{                     ATTextStream into AStream.                               }
{                     Removed ATRecordStream from TStream and created          }
{                     TFixedLenRecordStreamer.                                 }
{                     405 lines interface, 1307 lines implementation.          }
{                                                                              }


{                                                                              }
{ TStream                                                                      }
{   Base class for stream implementations.                                     }
{                                                                              }
{   TStream inherits from Delphi's TStream in the Classes unit to maintain     }
{   compatibality. Implementations must implement Read, Write, GetStreamSize,  }
{   SetStreamSize, GetPosition and SetPosition.                                }
{                                                                              }
{   EOF is True if no more octets *will* become available, if False more       }
{   octects *may* become available.                                            }
{                                                                              }
{   The OctetsAvailable property returns the number of octects that can be     }
{   read without blocking.                                                     }
{                                                                              }
{   Read reads from the stream, blocking until EOF or until the data becomes   }
{   available.                                                                 }
{                                                                              }
type
  EStream = class (Exception);
  TOctet = Byte;
  TEOLType = Set of (eol_CR, eol_LF, eol_CRLF, eol_LFCR, eol_EOF);
  TStream = class (Classes.TStream)
    { Classes.TStream compatibility                                            }
    protected
    Procedure SetSize (NewSize : Integer); override;
    Function  Seek (Offset : Integer; Origin : Word) : Integer; override;
    Function  Read (var Buffer; Count : Longint) : Longint; reintroduce; overload; override;
    Function  Write (const Buffer; Count : Longint) : Longint; reintroduce; overload; override;

    { Abstract methods                                                         }
    protected
    Function  GetStreamSize : Int64; virtual; abstract;
    Procedure SetStreamSize (const NewSize : Int64); virtual; abstract;
    Function  GetPosition : Int64; virtual; abstract;
    Procedure SetPosition (const NewPosition : Int64); virtual; abstract;

    public
    Function  Read : TOctet; overload; virtual;
    Function  Read (const Count : Integer) : String; overload; virtual;
    Procedure Write (const Data : TOctet); overload; virtual;
    Procedure Write (const Data : String); overload; virtual;
    Property  Size : Int64 read GetStreamSize write SetStreamSize;
    Property  Position : Int64 read GetPosition write SetPosition;
    Function  EOF : Boolean; virtual; abstract;
    Function  OctetsRemaining : Int64; virtual;
    Procedure WaitOctetAvailable (const TimeOut : Integer); virtual; abstract;
    Function  OctetsAvailable : Int64; virtual; abstract;

    { Other stream methods implemented using above abstract methods. Stream    }
    { implementations can override these with more efficient implementations.  }
    private
    Procedure CopyFrom (const S : Classes.TStream);

    protected
    Function  GetAsString : String; virtual;
    Procedure SetAsString (const S : String); virtual;

    public
    Property  AsString : String read GetAsString write SetAsString;
    Procedure Truncate; virtual;
    Procedure Skip; overload; virtual;
    Procedure Skip (const Count : Integer); overload; virtual;
    Function  Peek : TOctet; overload; virtual;
    Function  Peek (const Count : Integer) : String; overload; virtual;
    Procedure Assign (const S : Classes.TStream); virtual;
    Procedure Append; overload; virtual;
    Procedure Append (const S : Classes.TStream); overload; virtual;

    Function  Compare (const S : Classes.TStream; const Length : Int64) : Boolean; overload; virtual;
    Function  Compare (const S : String) : Boolean; overload; virtual;
    Function  Compare (const Pos, Length : Int64) : Boolean; overload; virtual;
    Function  IsEqual (const S : Classes.TStream) : Boolean; overload; virtual;
    Function  IsEqual (const S : String) : Boolean; overload; virtual;
    Function  Locate (const Find : String) : Int64; virtual;

    { Implements functionality to read/write binary data types from streams.   }
    Function  ReadPackedInteger : Integer; virtual;
    Function  ReadPackedString : String; virtual;
    Function  ReadPackedShortString : ShortString; virtual;
    Function  ReadPackedSingle : Single; virtual;
    Function  ReadPackedDouble : Double; virtual;
    Function  ReadPackedExtended : Extended; virtual;
    Function  ReadPackedCurrency : Currency; virtual;
    Function  ReadPackedBoolean : Boolean; virtual;
    Function  ReadPackedDateTime : TDateTime; virtual;
    Function  ReadPackedInt64 : Int64; virtual;
    Function  ReadPackedTInteger : TInteger;
    Function  ReadPackedTReal : TReal;

    Procedure WritePacked (const D : Integer); overload; virtual;
    Procedure WritePacked (const D : Extended); overload; virtual;
    Procedure WritePackedSingle (const D : Single); overload; virtual;
    Procedure WritePackedDouble (const D : Double); overload; virtual;
    Procedure WritePackedCurrency (const D : Currency); overload; virtual;
    Procedure WritePackedDateTime (const D : TDateTime); overload; virtual;
    Procedure WritePacked (const D : Boolean); overload; virtual;
    Procedure WritePacked (const D : Int64); overload; virtual;
    Procedure WritePacked (const D : String); overload; virtual;
    Procedure WritePackedShortString (const D : ShortString); overload; virtual;

    { Parsing functions                                                        }
    { The Skip functions return True if any characters were skipped; and exit  }
    {   without raising an exception if EOF is reached.                        }
    { SkipDelim skips up to *and over* Delimiter or EOF.                       }
    { SkipEOL skips over one EOL sequence.                                     }
    { Match functions return True if next characters match, without moving the }
    {   current position. If the match overruns EOF then the function returns  }
    {   False. MatchLine which will tolerate the FLineDelim past EOF.          }
    { MatchText is a case insensitive, locale sensitive match.                 }
    { The Extract functions return the extracted text, unchanged. Returns ''   }
    {   if a match could not be made.                                          }
    { All functions exit when EOF is reached without raising an exception,     }
    {   except Function Extract : Char.                                        }
    { ExtractText is a case insensitive, locale sensitive match.               }
    { ExtractRest extracts until EOF.                                          }
    { ExtractAll extracts while matching MatchChars.                           }
    { ExtractDelim extracts a string delimited by Delimiter or EOF and returns }
    {   the text up to but excluding the delimiter.                            }
    { ExtractLine is equivalent to ExtractDelim (FLineDelim); SkipLine;        }
    { ExtractKeyword uses MatchText to match from a list of possible keywords. }
    {   The keyword must be followed by DelimSet or EOF. Returns index of      }
    {   extracted keyword, or -1 if none.                                      }
    { ExtractQuotedText extracts text enclosed in a pair of the same from      }
    {   QuoteSet. Quotes can be included inside the text by using two          }
    {   consequetive quotes,                                                   }
    {   eg for QuoteSet = ['"', '''']  "A'BC"=A'BC, 'Dave''s'=Dave's, """"="   }
    protected
    FLineNr      : Integer;
    FLastNewLine : Integer;
    FEOLType     : TEOLType;

    Function  GetCol : Integer;
    Procedure DoReset;

    public
    Constructor Create;
    Procedure Reset; virtual;

    Function  SkipChar (const SkipSet : CharSet) : Boolean;
    Function  SkipChars (const SkipSet : CharSet) : Boolean;
    Function  SkipDelim (const Delim : String) : Boolean;
    Function  SkipLine : Boolean;
    Function  SkipWhiteSpace : Boolean;
    Function  SkipEOL : Boolean;

    Function  Match (const MatchChars : CharSet) : Boolean; overload;
    Function  Match (const Ch : Char) : Boolean; overload;
    Function  Match (const S : String) : Boolean; overload;
    Function  MatchText (const Txt : String) : Boolean;
    Function  MatchLine (const Txt : String) : Boolean;
    Function  MatchWhiteSpace : Boolean;
    Function  MatchEOL : Boolean;

    Function  Extract : Char; overload;
    Function  Extract (const MatchChars : CharSet) : String; overload;
    Function  Extract (const S : String) : String; overload;
    Function  ExtractText (const Txt : String) : String;
    Function  ExtractRest : String;

    Function  ExtractAll (const MatchChars : CharSet) : String; overload;
    Function  ExtractDelim (const Delimiter : CharSet; const SkipDelim : Boolean = False) : String; overload;
    Function  ExtractDelim (const Delimiter : String; const SkipDelim : Boolean = False) : String; overload;
    Function  ExtractLine : String;
    Function  ExtractKeyword (const Keywords : Array of String; const DelimSet : CharSet) : Integer;
    Function  ExtractQuotedText (const QuoteSet : CharSet) : String;
    Function  ExtractEscaped (const Escape : Char) : Char;

    Property  EOLType : TEOLType read FEOLType write FEOLType;
    Property  Line : Integer read FLineNr;
    Property  Col : Integer read GetCol;
  end;



{                                                                              }
{ TFixedLenRecordStreamer                                                      }
{   Implements functionality to access fixed-length records in streams.        }
{                                                                              }
type
  TFixedLenRecordStreamer = class
    protected
    FStream       : TStream;
    FRecordSize   : Integer;
    FRecordOffset : Int64;

    Function  GetRecordCount : Int64;
    Procedure SetRecordCount (const NewRecordCount : Int64);
    Procedure SetRecordSize (const RecordSize : Integer);
    Procedure Seek (const Idx : Integer);

    public
    Constructor Create (const Stream : TStream; const RecordSize : Integer = 1;
                const FirstRecordOffset : Int64 = 0);
    Destructor Destroy; override;

    Procedure Read (const Idx, Count : Integer; var Buf);
    Procedure Write (const Idx, Count : Integer; const Buf);
    Procedure Copy (const SrcIdx, DestIdx, Count : Integer);
    Procedure Insert (const Idx, Count : Integer; const Buf);
    Procedure Delete (const Idx, Count : Integer);

    Property RecordSize : Integer read FRecordSize write FRecordSize;
    Property RecordOffset : Int64 read FRecordOffset write FRecordOffset;
    Property RecordCount : Int64 read GetRecordCount write SetRecordCount;
  end;



{                                                                              }
{ TFileStream                                                                  }
{   A file implemented as a stream.                                            }
{                                                                              }
type
  TFileStream = class (TStream)
    protected
    FFileStream : Classes.TFileStream;

    Function GetStreamSize : Int64; override;
    Procedure SetStreamSize (const NewSize : Int64); override;

    public
    Constructor Create (const FilePath : String; const CreateFile : Boolean); reintroduce; overload;
    Destructor Destroy; override;

    Procedure Skip (const Count : Integer); overload; override;
    Procedure Skip; overload; override;
    Function  Read : TOctet; overload; override;
    Function  Read (const Count : Integer) : String; overload; override;
    Function  Read (var Buf; Count : Integer) : Integer; overload; override;
    Function  Peek : TOctet; overload; override;
    Function  Peek (const Count : Integer) : String; overload; override;
    Function  GetPosition : Int64; override;
    Procedure SetPosition (const NewPosition : Int64); override;
    Procedure Append; override;
    Function  OctetsRemaining : Int64; override;
    Function  EOF : Boolean; override;
    Function  OctetsAvailable : Int64; override;
    Procedure Truncate; override;
    Function  Write (const Buf; Count : Integer) : LongInt; overload; override;
    Procedure Write (const Data : TOctet); overload; override;
    Procedure Write (const Data : String); overload; override;
    Function  Locate (const Find : String) : Int64; override;
  end;



{                                                                              }
{ TStringStream                                                                }
{   A implementation of a stream that stores/retrieves data using a Delphi     }
{   String.                                                                    }
{                                                                              }
type
  TStringStream = class (TStream)
    protected
    FData : String;
    FPos  : Integer;

    Function GetStreamSize : Int64; override;
    Procedure SetStreamSize (const NewSize : Int64); override;

    public
    Constructor Create (const S : String); reintroduce; overload;

    Procedure Skip (const Count : Integer); overload; override;
    Procedure Skip; overload; override;
    Function  Read : TOctet; overload; override;
    Function  Read (const Count : Integer) : String; overload; override;
    Function  Read (var Buf; Count : Integer) : Integer; override;
    Function  Peek : TOctet; overload; override;
    Function  Peek (const Count : Integer) : String; overload; override;
    Function  GetPosition : Int64; override;
    Procedure SetPosition (const NewPosition : Int64); override;
    Procedure Append; override;
    Function  OctetsRemaining : Int64; override;
    Function  EOF : Boolean; override;
    Function  OctetsAvailable : Int64; override;
    Procedure Truncate; override;
    Function  Write (const Buf; Count : Integer) : LongInt; overload; override;
    Procedure Write (const Data : TOctet); overload; override;
    Procedure Write (const Data : String); overload; override;
    Function  Locate (const Find : String) : Int64; override;
    Procedure WaitOctetAvailable (const TimeOut : Integer); override;
  end;



{                                                                              }
{ TSocketStream                                                                }
{                                                                              }
type
  TSocketStream = class (TStream)
    protected
    FSocket     : TCustomWinSocket;
    FSockStream : TWinSocketStream;
    FBytesRead  : Integer;
    FPeekBuf    : String;
    FTimeOut    : Integer;

    Procedure DoReset;
    Procedure FillPeekBuf (const Len : Integer);

    public
    Constructor Create (const Socket : TCustomWinSocket; const TimeOut : Integer);
    Destructor Destroy; override;

    Procedure Skip (const Count : Integer); overload; override;
    Procedure Skip; overload; override;
    Function  Read (var Buf; Count : Integer) : Integer; override;
    Function  Read : TOctet; overload; override;
    Function  Read (const Count : Integer) : String; overload; override;
    Function  Peek : TOctet; overload; override;
    Function  Peek (const Count : Integer) : String; overload; override;
    Procedure Reset; override;
    Function  GetPosition : Int64; override;
    Procedure SetPosition (const NewPosition : Int64); override;
    Procedure Append; override;
    Function  EOF : Boolean; override;
    Function  OctetsAvailable : Int64; override;
    Procedure Write (const Data : TOctet); overload; override;
    Procedure Write (const Data : String); overload; override;
    Function  Write (const Buf; Count : Integer) : LongInt; overload; override;
    Procedure WaitOctetAvailable (const TimeOut : Integer); override;
  end;



implementation



uses
  Forms,
  
  // System units
  Windows,              // GetTickCount
  Math,                 // Max

  // Delphi Fundamentals (L0)
  cStrings;             // Pack


{                                                                              }
{ TStream                                                                      }
{                                                                              }
Function TStream.Read (var Buffer; Count : Longint) : Longint;
  Begin
    raise EAbstractError.Create ('Method TStream.Read not implemented.');
  End;

Function TStream.Read : TOctet;
  Begin
    if Read (Result, Sizeof (Result)) <> Sizeof (Result) then
      raise EReadError.Create ('Stream read error.');
  End;

Function TStream.Read (const Count : Integer) : String;
  Begin
    SetLength (Result, Count);
    if Count > 0 then
      if Read (Result [1], Count) <> Count then
        raise EReadError.Create ('Stream read error.');
  End;

Function TStream.Write (const Buffer; Count : Longint) : Longint;
  Begin
    raise EAbstractError.Create ('Method TStream.Write not implemented.');
  End;

Procedure TStream.Write (const Data : TOctet);
  Begin
    if Write (Data, Sizeof (Data)) <> Sizeof (Data) then
      raise EWriteError.Create ('Stream write error.');
  End;

Procedure TStream.Write (const Data : String);
var L : Integer;
  Begin
    L := Length (Data);
    if L > 0 then
      if Write (Data [1], L) <> L then
        raise EWriteError.Create ('Stream write error.');
  End;

Function TStream.Seek (Offset : Integer; Origin : Word) : Integer;
  Begin
    Case Origin of
      soFromBeginning : if Offset >= 0 then
                          SetPosition (Offset) else
                          raise EStream.Create ('Seek before bof');
      soFromCurrent   : SetPosition (GetPosition + Offset);
      soFromEnd       : SetPosition (GetStreamSize + Offset);
    end;
    Result := GetPosition;
  End;

Procedure TStream.SetSize (NewSize : Integer);
  Begin
    SetStreamSize (NewSize);
  End;

Function TStream.OctetsRemaining : Int64;
  Begin
    Result := Size - Position;
  End;

Procedure TStream.Truncate;
  Begin
    Size := Position;
  End;

Function TStream.GetAsString : String;
const ReadChunkSize = 4096;
  Begin
    Position := 0;
    While not EOF do
      Result := Result + Read (ReadChunkSize);
  End;

Procedure TStream.SetAsString (const S : String);
  Begin
    Position := 0;
    Write (S);
    Truncate;
  End;

Procedure TStream.Skip;
  Begin
    Position := Position + 1;
  End;

Procedure TStream.Skip (const Count : Integer);
  Begin
    Position := Min (Size, Position + Count);
  End;

Function TStream.Peek : TOctet;
var P : Int64;
  Begin
    P := Position;
    Result := Read;
    Position := P;
  End;

Function TStream.Peek (const Count : Integer) : String;
var P : Int64;
  Begin
    P := Position;
    Result := Read (Count);
    Position := P;
  End;

Procedure TStream.CopyFrom (const S : Classes.TStream);
const CopyChunkSize = 16384;
var Buf     : String;
    F, P, L : Int64;
  Begin
    S.Seek (0, soFromBeginning);
    P := 0;
    L := S.Size;
    Repeat
      F := Min (CopyChunkSize, L - P);
      if F > 0 then
        begin
          SetLength (Buf, F);
          S.ReadBuffer (Buf [1], F);
          Write (Buf);
          Inc (P, F);
        end;
    Until F = 0;
  End;

Procedure TStream.Assign (const S : Classes.TStream);
  Begin
    Position := 0;
    CopyFrom (S);
    Truncate;
  End;

Procedure TStream.Append;
  Begin
    Position := Size;
  End;

Procedure TStream.Append (const S : Classes.TStream);
  Begin
    Append;
    CopyFrom (S);
  End;

{ Compare functions leave stream at pre-search position if not found,          }
{ otherwise, at position after match.                                          }
Function TStream.Compare (const S : Classes.TStream; const Length : Int64) : Boolean;
const CompareChunkSize = 1024;
var P, I, R : Int64;
    Buf     : String;
  Begin
    Result := False;
    if (OctetsRemaining < Length) or (S.Size - S.Position < Length) then
      exit;

    P := Position;
    R := Length;
    While R > 0 do
      begin
        I := Min (R, CompareChunkSize);
        SetLength (Buf, I);
        S.ReadBuffer (Buf [1], I);
        if Read (I) <> Buf then
          begin
            Position := P;
            exit;
          end;
        Dec (R, I);
      end;
    Result := True;
  End;

Function TStream.Compare (const S : String) : Boolean;
const CompareChunkSize = 1024;
var I, R : Integer;
    P : Int64;
  Begin
    Result := False;
    if OctetsRemaining < Length (S) then
      exit;

    P := Position;
    R := Length (S);
    While R > 0 do
      begin
        I := Min (R, CompareChunkSize);
        if Read (I) <> Copy (S, Length (S) - R + 1, I) then
          begin
            Position := P;
            exit;
          end;
        Dec (R, I);
      end;
    Result := True;
  End;

Function TStream.IsEqual (const S : Classes.TStream) : Boolean;
var L1, L2 : Int64;
  Begin
    L1 := Size;
    L2 := S.Size;
    if (L1 >= 0) and (L2 >= 0) and (L1 <> L2) then
      Result := False else
      begin
        Position := 0;
        S.Seek (0, soFromBeginning);
        Result := Compare (S, L1);
      end;
  End;

Function TStream.IsEqual (const S : String) : Boolean;
  Begin
    if Length (S) <> OctetsRemaining then
      Result := False else
      begin
        Position := 0;
        Result := Compare (S);
      end;
  End;

Function TStream.Compare (const Pos, Length : Int64) : Boolean;
const CompareChunkSize = 1024;
var I, R, P : Int64;
    S       : String;
  Begin
    Result := False;
    if OctetsRemaining < Length then
      exit;

    R := Length;
    P := Position;
    While R > 0 do
      begin
        I := Min (R, CompareChunkSize);
        S := Read (I);
        Position := Pos + Length - R;
        if S <> Read (I) then
          begin
            Position := P;
            exit;
          end;
        Dec (R, I);
        Position := P + Length - R;
      end;
    Result := True;
  End;

{ Locate returns the position (relative to the current position) of the        }
{ first occurance of Find in the stream. It does not change the current        }
{ position. Returns -1 if Find could not be found.                             }
Function TStream.Locate (const Find : String) : Int64;
var P, I : Int64;
  Begin
    P := Position;
    I := 0;
    Repeat
      if Compare (Find) then
        begin
          Position := P;
          Result := I;
          exit;
        end;
      Inc (I);
      Position := P + I;
    Until EOF;
    Position := P;
    Result := -1;
  End;

{                                                                              }
{ ATBinaryStream                                                               }
{                                                                              }
{   Uses internal Delphi formats:                                              }
{     Integer       4 byte integer (LSB first)                                 }
{     Int64         8 byte integer                                             }
{     Single        4 byte FPU Single                                          }
{     Double        8 byte FPU Double                                          }
{     Extended      10 byte FPU Extended                                       }
{     Currency      8 byte FPU fixed point                                     }
{     Boolean       1 byte, 0 = False                                          }
{     DateTime      8 byte FPU Double                                          }
{     String        Integer length + String                                    }
{     ShortString   Byte length + String                                       }
{                                                                              }
Function TStream.ReadPackedInteger : Integer;
  Begin
    Move (Read (Sizeof (Result)) [1], Result, Sizeof (Result));
  End;

Function TStream.ReadPackedInt64 : Int64;
  Begin
    Move (Read (Sizeof (Result)) [1], Result, Sizeof (Result));
  End;

Function TStream.ReadPackedSingle : Single;
  Begin
    Move (Read (Sizeof (Result)) [1], Result, Sizeof (Result));
  End;

Function TStream.ReadPackedDouble : Double;
  Begin
    Move (Read (Sizeof (Result)) [1], Result, Sizeof (Result));
  End;

Function TStream.ReadPackedExtended : Extended;
  Begin
    Move (Read (Sizeof (Result)) [1], Result, Sizeof (Result));
  End;

Function TStream.ReadPackedCurrency : Currency;
  Begin
    Move (Read (Sizeof (Result)) [1], Result, Sizeof (Result));
  End;

Function TStream.ReadPackedBoolean : Boolean;
  Begin
    Move (Read (Sizeof (Result)) [1], Result, Sizeof (Result));
  End;

Function TStream.ReadPackedDateTime : TDateTime;
  Begin
    Move (Read (Sizeof (Result)) [1], Result, Sizeof (Result));
  End;

Function TStream.ReadPackedTInteger : TInteger;
  Begin
    Move (Read (Sizeof (Result)) [1], Result, Sizeof (Result));
  End;

Function TStream.ReadPackedTReal : TReal;
  Begin
    Move (Read (Sizeof (Result)) [1], Result, Sizeof (Result));
  End;

Function TStream.ReadPackedString : String;
var L : Integer;
  Begin
    L := ReadPackedInteger;
    SetLength (Result, L);
    if L > 0 then
      Move (Read (L) [1], Result [1], L);
  End;

Function TStream.ReadPackedShortString : ShortString;
var L : Byte;
  Begin
    L := Read;
    SetLength (Result, L);
    if L > 0 then
      Move (Read (L) [1], Result [1], L);
  End;

Procedure TStream.WritePacked (const D : Integer);
  Begin
    Write (Pack (D));
  End;

Procedure TStream.WritePacked (const D : Int64);
  Begin
    Write (Pack (D));
  End;

Procedure TStream.WritePacked (const D : String);
  Begin
    Write (Pack (D));
  End;

Procedure TStream.WritePackedShortString (const D : ShortString);
  Begin
    Write (PackShortString (D));
  End;

Procedure TStream.WritePackedSingle (const D : Single);
  Begin
    Write (PackSingle (D));
  End;

Procedure TStream.WritePackedDouble (const D : Double);
  Begin
    Write (PackDouble (D));
  End;

Procedure TStream.WritePacked (const D : Extended);
  Begin
    Write (Pack (D));
  End;

Procedure TStream.WritePackedCurrency (const D : Currency);
  Begin
    Write (PackCurrency (D));
  End;

Procedure TStream.WritePacked (const D : Boolean);
  Begin
    Write (Pack (D));
  End;

Procedure TStream.WritePackedDateTime (const D : TDateTime);
  Begin
    Write (PackDateTime (D));
  End;

{                                                                              }
{ ATTextStream                                                                 }
{                                                                              }
Procedure TStream.DoReset;
  Begin
    FLineNr := 1;
    FLastNewLine := 0;
  End;

Constructor TStream.Create;
  Begin
    inherited Create;
    DoReset;
    FEOLType := [eol_CRLF, eol_EOF];
  End;

Function TStream.GetCol : Integer;
  Begin
    Result := GetPosition - FLastNewLine + 1;
  End;

Procedure TStream.Reset;
  Begin
    DoReset;
    Position := 0;
  End;

{ Skip                                                                         }
Function TStream.SkipChar (const SkipSet : CharSet) : Boolean;
  Begin
    Result := not EOF and (Char (Peek) in SkipSet);
    if Result then
      Skip;
  End;

Function TStream.SkipChars (const SkipSet : CharSet) : Boolean;
  Begin
    Result := SkipChar (SkipSet);
    if Result then
      Repeat
      Until not SkipChar (SkipSet);
  End;

Function TStream.SkipDelim (const Delim : String) : Boolean;
var R : Boolean;
  Begin
    Result := False;
    if Length (Delim) = 0 then
      exit;

    Repeat
      While not EOF and not (Char (Peek) = Delim [1]) do
        begin
          Skip;
          Result := True;
        end;
      if EOF then
        exit;

      R := cStrings.Match (Delim, Peek (Length (Delim)));
      if not R then
        Skip;
    Until R;
    Result := True;
    Skip (Length (Delim));
  End;

Function TStream.SkipWhiteSpace : Boolean;
  Begin
    Result := SkipChars (cs_WhiteSpace)
  End;

Function TStream.SkipLine : Boolean;
  Begin
    While not MatchEOL do
      Skip;
    Result := SkipEOL;
  End;

Function TStream.SkipEOL : Boolean;
var Ch     : Char;
    CR, LF : Boolean;
    S      : String;
  Begin
    Result := (eol_EOF in FEOLType) and EOF;
    if not Result then
      begin
        CR := eol_CRLF in FEOLType;
        LF := eol_LFCR in FEOLType;
        if CR or LF then
          begin
            S := Read (2);
            Result := CR and (S = ASCII_CR + ASCII_LF) or
                      LF and (S = ASCII_LF + ASCII_CR);
            if not Result then
              begin
                Result := (eol_CR in FEOLType) and (S [1] = ASCII_CR) or
                          (eol_LF in FEOLType) and (S [1] = ASCII_LF);
                if Result then
                  Position := Position - 1 else
                  Position := Position - 2;
              end;
          end else
          begin
            CR := eol_CR in FEOLType;
            LF := eol_LF in FEOLType;
            if CR or LF then
              begin
                Ch := Char (Read);
                Result := CR and (Ch = ASCII_CR) or
                          LF and (Ch = ASCII_LF);
                if not Result then
                  Position := Position - 1;
              end;
          end;
      end;

    if Result then
      begin
        Inc (FLineNr);
        FLastNewLine := GetPosition;
      end;
  End;

{ Match                                                                        }
Function TStream.MatchEOL : Boolean;
var CR, LF : Boolean;
    Ch     : Char;
    S      : String;
  Begin
    Result := (eol_EOF in FEOLType) and EOF;
    if not Result then
      begin
        CR := eol_CR in FEOLType;
        LF := eol_LF in FEOLType;
        if CR or LF then
          begin
            Ch := Char (Peek);
            Result := CR and (Ch = ASCII_CR) or
                      LF and (Ch = ASCII_LF);
          end;

        if not Result then
          begin
            CR := eol_CRLF in FEOLType;
            LF := eol_LFCR in FEOLType;
            if CR or LF then
              begin
                S := Peek (2);
                Result := CR and (S = ASCII_CR + ASCII_LF) or
                          LF and (S = ASCII_LF + ASCII_CR);
              end;
          end;
      end;
  End;

Function TStream.Match (const MatchChars : CharSet) : Boolean;
  Begin
    Result := not EOF and (Char (Peek) in MatchChars);
  End;

Function TStream.MatchWhiteSpace : Boolean;
  Begin
    Result := Match (cs_WhiteSpace);
  End;

Function TStream.Match (const Ch : Char) : Boolean;
  Begin
    Result := not EOF and (Char (Peek) = Ch);
  End;

Function TStream.Match (const S : String) : Boolean;
var L, R : Integer;
  Begin
    L := Length (S);
    R := OctetsRemaining;
    if (R >= 0) and (R < L) then
      begin
        Result := False;
        exit;
      end;

    Result := Peek (L) = S;
  End;

Function TStream.MatchText (const Txt : String) : Boolean;
var L, R : Integer;
  Begin
    L := Length (Txt);
    R := OctetsRemaining;
    if (R >= 0) and (R < L) then
      begin
        Result := False;
        exit;
      end;

    Result := AnsiCompareText (Txt, Peek (L)) = 0;
  End;

Function TStream.MatchLine (const Txt : String) : Boolean;
var S : String;
  Begin
    S := Extract (Txt);
    Result := (S = Txt) and MatchEOL;
    if Result then
      SkipEOL;
  End;

{ Extract                                                                      }
Function TStream.Extract (const MatchChars : CharSet) : String;
  Begin
    if Match (MatchChars) then
      Result := Extract else
      Result := '';
  End;

Function TStream.Extract : Char;
  Begin
    Result := Char (Read);
  End;

Function TStream.Extract (const S : String) : String;
  Begin
    if Match (S) then
      begin
        Skip (Length (S));
        Result := S;
      end else
      Result := '';
  End;

Function TStream.ExtractText (const Txt : String) : String;
  Begin
    if MatchText (Txt) then
      Result := Read (Length (Txt)) else
      Result := '';
  End;

Function TStream.ExtractRest : String;
  Begin
    Result := '';
    While not EOF do
      Result := Result + Read (Max (1, OctetsAvailable));
  End;

{ ExtractDelim                                                                 }
Function TStream.ExtractAll (const MatchChars : CharSet) : String;
  Begin
    Result := '';
    While not EOF and (Char (Peek) in MatchChars) do
      Result := Result + Char (Read);
  End;

Function TStream.ExtractDelim (const Delimiter : CharSet; const SkipDelim : Boolean = False) : String;
  Begin
    Result := '';
    While not EOF and not (Char (Peek) in Delimiter) do
      Result := Result + Char (Read);
    if SkipDelim then
      Skip;
  End;

Function TStream.ExtractDelim (const Delimiter : String; const SkipDelim : Boolean = False) : String;
var R : Boolean;
  Begin
    Result := '';
    R := False;
    While not EOF and not R do
      begin
        R := Peek (Length (Delimiter)) = Delimiter;
        if not R then
          Result := Result + Char (Read);
      end;
    if SkipDelim and R then
      Skip (Length (Delimiter));
  End;

Function TStream.ExtractEscaped (const Escape : Char) : Char;
  Begin
    if Match (Escape) then
      Skip;
    Result := Extract;
  End;

Function TStream.ExtractQuotedText (const QuoteSet : CharSet) : String;
var Quote : Char;
    Fin   : Boolean;
  Begin
    if Match (QuoteSet) then
      begin
        Quote := Extract;
        Result := '';
        Repeat
          Result := Result + ExtractDelim ([Quote]);
          if EOF then
            raise EStream.Create ('No closing quote (' + Quote + ' expected)');
          Skip;
          Fin := not Match (Quote);
          if not Fin then
            begin
              Result := Result + Quote;
              Skip;
            end;
        Until Fin;
      end else
      Result := '';
  End;

Function TStream.ExtractLine : String;
  Begin
    Result := '';
    While not MatchEOL do
      Result := Result + Extract;
    SkipEOL;
  End;

Function TStream.ExtractKeyword (const Keywords : Array of String; const DelimSet : CharSet) : Integer;
var I : Integer;
    R : Boolean;
    S : String;
  Begin
    For I := 0 to High (Keywords) do
      if MatchText (Keywords [I]) then
        begin
          S := Peek (Length (Keywords [I]) + 1);
          R := (Length (S) = Length (Keywords [I])) or (DelimSet = []);
          if R or (S [Length (S)] in DelimSet) then
            begin
              Result := I;
              Skip (Length (Keywords [I]));
              if not R then
                Skip;
              exit;
            end;
        end;
    Result := -1;
  End;


{                                                                              }
{ TFixedLenRecordStreamer                                                      }
{                                                                              }
Constructor TFixedLenRecordStreamer.Create (const Stream : TStream; const RecordSize : Integer; const FirstRecordOffset : Int64);
  Begin
    inherited Create;
    FStream := Stream;
    FRecordSize := RecordSize;
    FRecordOffset := FirstRecordOffset;
  End;

Destructor TFixedLenRecordStreamer.Destroy;
  Begin
    FreeAndNil (FStream);
    inherited Destroy;
  End;

Function TFixedLenRecordStreamer.GetRecordCount : Int64;
  Begin
    Result := (FStream.Size - FRecordOffset) div FRecordSize;
  End;

Procedure TFixedLenRecordStreamer.SetRecordCount (const NewRecordCount : Int64);
  Begin
    FStream.SetStreamSize (FRecordOffset + NewRecordCount * FRecordSize);
  End;

Procedure TFixedLenRecordStreamer.SetRecordSize (const RecordSize : Integer);
  Begin
    if RecordSize <= 0 then
      raise EStream.Create ('Invalid record size') else
      FRecordSize := RecordSize;
  End;

Procedure TFixedLenRecordStreamer.Seek (const Idx : Integer);
  Begin
    FStream.Position := FRecordOffset + Idx * FRecordSize;
  End;

Procedure TFixedLenRecordStreamer.Read (const Idx, Count : Integer; var Buf);
  Begin
    Seek (Idx);
    FStream.Read (Buf, Count * FRecordSize);
  End;

Procedure TFixedLenRecordStreamer.Write (const Idx, Count : Integer; const Buf);
  Begin
    Seek (Idx);
    FStream.Write (Buf, Count * FRecordSize);
  End;

Procedure TFixedLenRecordStreamer.Copy (const SrcIdx, DestIdx, Count : Integer);
var Buf : String;
  Begin
    SetLength (Buf, Count * FRecordSize);
    Read (SrcIdx, Count, Buf [1]);
    Write (DestIdx, Count, Buf);
  End;

Procedure TFixedLenRecordStreamer.Insert (const Idx, Count : Integer; const Buf);
  Begin
    Copy (Idx, Idx + Count, RecordCount - Idx);
    Write (Idx, Count, Buf);
  End;

Procedure TFixedLenRecordStreamer.Delete (const Idx, Count : Integer);
  Begin
    Copy (Idx + Count, Idx, RecordCount - Idx - Count);
    FStream.Position := (RecordCount - Count) * FRecordSize;
    FStream.Truncate;
  End;



{                                                                              }
{ TFileStream                                                                  }
{                                                                              }
Constructor TFileStream.Create (const FilePath : String; const CreateFile : Boolean);
  Begin
    inherited Create;
    if CreateFile then
      FFileStream := Classes.TFileStream.Create (FilePath, fmCreate) else
      FFileStream := Classes.TFileStream.Create (FilePath, fmOpenReadWrite or fmShareCompat or fmShareDenyNone);
  End;

Destructor TFileStream.Destroy;
  Begin
    FreeAndNil (FFileStream);
    inherited Destroy;
  End;

Function TFileStream.Read : TOctet;
  Begin
    FFileStream.ReadBuffer (Result, Sizeof (Result));
  End;

Function TFileStream.Read (var Buf; Count : Integer) : Integer;
  Begin
    if Count <= 0 then
      Result := 0 else
      Result := FFileStream.Read (Buf, Min (Count, OctetsRemaining));
  End;

Function TFileStream.Read (const Count : Integer) : String;
var I : Integer;
  Begin
    if Count <= 0 then
      Result := '' else
      begin
        I := Min (Count, OctetsRemaining);
        SetLength (Result, I);
        if I > 0 then
          SetLength (Result, FFileStream.Read (Result [1], I));
      end;
  End;

Procedure TFileStream.Write (const Data : TOctet);
  Begin
    FFileStream.WriteBuffer (Data, Sizeof (Data));
  End;

Function TFileStream.Write (const Buf; Count : Integer) : LongInt;
  Begin
    FFileStream.WriteBuffer (Buf, Count);
    Result := Count;
  End;

Procedure TFileStream.Write (const Data : String);
  Begin
    FFileStream.WriteBuffer (Data [1], Length (Data));
  End;

Function TFileStream.OctetsAvailable : Int64;
  Begin
    Result := FFileStream.Size - FFileStream.Position;
  End;

Function TFileStream.EOF : Boolean;
  Begin
    Result := not (FFileStream.Position < FFileStream.Size);
  End;

Procedure TFileStream.Skip;
  Begin
    if EOF then
      raise EStream.Create ('Skip past EOF');
    FFileStream.Seek (FFileStream.Position + 1, soFromBeginning);
  End;

Procedure TFileStream.Skip (const Count : Integer);
  Begin
    FFileStream.Seek (Min (FFileStream.Position + Count, FFileStream.Size), soFromBeginning);
  End;

Function TFileStream.Peek : TOctet;
var P : Integer;
  Begin
    P := FFileStream.Position;
    FFileStream.Read (Result, Sizeof (Result));
    FFileStream.Seek (P, soFromBeginning);
  End;

Function TFileStream.Peek (const Count : Integer) : String;
var P : Integer;
  Begin
    P := FFileStream.Position;
    Result := Read (Count);
    FFileStream.Seek (P, soFromBeginning);
  End;

Function TFileStream.GetPosition : Int64;
  Begin
    Result := FFileStream.Position;
  End;

Function TFileStream.GetStreamSize : Int64;
  Begin
    Result := FFileStream.Size;
  End;

Procedure TFileStream.SetStreamSize (const NewSize : Int64);
  Begin
    FFileStream.Size := NewSize;
  End;

Procedure TFileStream.SetPosition (const NewPosition : Int64);
  Begin
    FFileStream.Position := NewPosition;
  End;

Procedure TFileStream.Truncate;
  Begin
    FFileStream.Size := FFileStream.Position;
  End;

Procedure TFileStream.Append;
  Begin
    FFileStream.Seek (0, soFromEnd);
  End;

Function TFileStream.OctetsRemaining : Int64;
  Begin
    Result := FFileStream.Size - FFileStream.Position;
  End;

Function TFileStream.Locate (const Find : String) : Int64;

  Function ReadBuf (const Len : Integer) : String;
    Begin
      SetLength (Result, Len);
      SetLength (Result, FFileStream.Read (Result [1], Len));
    End;

const BufSize = 1024;

var Buf, B : String;
    P, I   : Integer;

  Begin
    P := FFileStream.Position;
    try
      Buf := ReadBuf (Length (Find));
      Result := Pos (Find, Buf);
      if Result > 0 then
        exit;
      I := 0;
      While FFileStream.Position < FFileStream.Size do
        begin
          B := ReadBuf (BufSize);
          Buf := Buf + B;
          Result := I + Pos (Find, Buf);
          if Result > I then
            exit;
          System.Delete (Buf, 1, Length (B));
          Inc (I, Length (B));
        end;
      Result := 0;
    finally
      FFileStream.Seek (P, soFromBeginning);
    end;
  End;



{                                                                              }
{ TStringStream                                                                }
{                                                                              }
Constructor TStringStream.Create (const S : String);
  Begin
    inherited Create;
    FData := S;
  End;

Function TStringStream.Read : TOctet;
  Begin
    if FPos = Length (FData) then
      raise EStream.Create ('Read past EOF');
    Inc (FPos);
    Result := TOctet (FData [FPos]);
  End;

Function TStringStream.Read (var Buf; Count : Integer) : Integer;
  Begin
    if FPos + Count > Length (FData) then
      Result := Length (FData) - FPos else
      Result := Count;
    Move (FData [FPos + 1], Buf, Result);
    Inc (FPos, Result);
  End;

Function TStringStream.Read (const Count : Integer) : String;
var L : Integer;
  Begin
    if FPos + Count > Length (FData) then
      L := Length (FData) - FPos else
      L := Count;
    Result := Copy (FData, FPos + 1, L);
    Inc (FPos, L);
  End;

Procedure TStringStream.Write (const Data : TOctet);
  Begin
    if FPos = Length (FData) then
      FData := FData + Char (Data) else
      FData [FPos] := Char (Data);
    Inc (FPos);
  End;

Function TStringStream.Write (const Buf; Count : Integer) : LongInt;
  Begin
    if FPos + Count > Length (FData) then
      SetLength (FData, FPos + Count);
    Move (Buf, FData [FPos + 1], Count);
    Inc (FPos, Count);
    Result := Count;
  End;

Procedure TStringStream.Write (const Data : String);
  Begin
    if FPos + Length (Data) > Length (FData) then
      SetLength (FData, FPos + Length (Data));
    Move (Data [1], FData [FPos + 1], Length (Data));
    Inc (FPos, Length (Data));
  End;

Function TStringStream.EOF : Boolean;
  Begin
    Result := FPos >= Length (FData);
  End;

Procedure TStringStream.Skip;
  Begin
    if EOF then
      raise EStream.Create ('Skip past EOF');
    Inc (FPos);
  End;

Procedure TStringStream.Skip (const Count : Integer);
  Begin
    if FPos + Count > Length (FData) then
      FPos := Length (FData) else
      Inc (FPos, Count);
  End;

Function TStringStream.Peek : TOctet;
  Begin
    if EOF then
      raise EStream.Create ('Peek past EOF');
    Result := TOctet (FData [FPos + 1]);
  End;

Function TStringStream.Peek (const Count : Integer) : String;
  Begin
    Result := Copy (FData, FPos + 1, Count);
  End;

Function TStringStream.GetPosition : Int64;
  Begin
    Result := FPos;
  End;

Procedure TStringStream.SetPosition (const NewPosition : Int64);
  Begin
    if NewPosition > Length (FData) then
      raise EStream.Create ('Seek past EOF');
    FPos := NewPosition;
  End;

Procedure TStringStream.Append;
  Begin
    FPos := Length (FData);
  End;

Function TStringStream.OctetsRemaining : Int64;
  Begin
    Result := Length (FData) - FPos;
  End;

Function TStringStream.OctetsAvailable : Int64;
  Begin
    Result := Length (FData) - FPos;
  End;

Function TStringStream.GetStreamSize : Int64;
  Begin
    Result := Length (FData);
  End;

Procedure TStringStream.SetStreamSize (const NewSize : Int64);
  Begin
    SetLength (FData, NewSize);
  End;

Procedure TStringStream.Truncate;
  Begin
    SetLength (FData, FPos);
  End;

Function TStringStream.Locate (const Find : String) : Int64;
  Begin
    Result := Pos (Find, Copy (FData, FPos + 1, Length (FData) - FPos));
  End;

Procedure TStringStream.WaitOctetAvailable (const TimeOut : Integer);
  Begin
  End;

{ TSocketStream                                                                }
Constructor TSocketStream.Create (const Socket : TCustomWinSocket; const TimeOut : Integer);
  Begin
    inherited Create;
    FSocket := Socket;
    FTimeOut := TimeOut;
    FSockStream := TWinSocketStream.Create (FSocket, TimeOut);
    DoReset;
  End;

Destructor TSocketStream.Destroy;
  Begin
    FreeAndNil (FSockStream);
    inherited Destroy;
  End;


Procedure TSocketStream.DoReset;
  Begin
    FBytesRead := 0;
    FPeekBuf := '';
  End;

Function TSocketStream.EOF : Boolean;
  Begin
    Result := not FSocket.Connected;
  End;

procedure TSocketStream.Append;
  Begin
  End;

Function TSocketStream.OctetsAvailable : Int64;
  Begin
    Result := Length (FPeekBuf) + FSocket.ReceiveLength;
  End;

Procedure TSocketStream.FillPeekBuf (const Len : Integer);
var L : Integer;
    S : String;
  Begin
    L := Len - Length (FPeekBuf);
    if L <= 0 then
      exit;
    FSockStream.WaitForData (FTimeOut);
    SetLength (S, L);
    SetLength (S, FSockStream.Read (S [1], L));
    if S <> '' then
      FPeekBuf := FPeekBuf + S;
  End;

Procedure TSocketStream.WaitOctetAvailable (const TimeOut : Integer);
  Begin
    FillPeekBuf (Sizeof (TOctet));
  End;

Function TSocketStream.Peek : TOctet;
  Begin
    FillPeekBuf (Sizeof (Result));
    Result := TOctet (FPeekBuf [1]);
  End;

Function TSocketStream.Peek (const Count : Integer) : String;
  Begin
    FillPeekBuf (Count);
    Result := CopyLeft (FPeekBuf, Count);
  End;

Function TSocketStream.Read (var Buf; Count : Integer) : Integer;
type ByteBuffer = Array [0..2140000000] of Byte;
var Buffer : ByteBuffer absolute Buf;
    I : Integer;
  Begin
    if FPeekBuf <> '' then
      begin
        I := Min (Count, Length (FPeekBuf));
        Move (FPeekBuf [1], Buffer [1], I);
        Delete (FPeekBuf, 1, I);
        Dec (Count, I);
      end else
      I := 0;
    Result := I + FSockStream.Read (Buffer [I], Count);
  End;

Function TSocketStream.Read (const Count : Integer) : String;
  Begin
    FillPeekBuf (Count);
    Result := CopyLeft (FPeekBuf, Count);
    System.Delete (FPeekBuf, 1, Count);
    Inc (FBytesRead, Count);
  End;

Function TSocketStream.Read : TOctet;
  Begin
    FillPeekBuf (Sizeof (Result));
    Result := TOctet (FPeekBuf [1]);
    System.Delete (FPeekBuf, 1, 1);
    Inc (FBytesRead);
  End;

Procedure TSocketStream.Reset;
  Begin
    DoReset;
    inherited Reset;
  End;

Function TSocketStream.GetPosition : Int64;
  Begin
    Result := FBytesRead;
  End;

Procedure TSocketStream.SetPosition (const NewPosition : Int64);
var P : Int64;
  Begin
    P := Position;
    if NewPosition >= P then
      Read (NewPosition - P) else
      raise EStream.Create ('TSocketStream does not support seeking backwards.');
  End;

Procedure TSocketStream.Skip;
  Begin
    Read;
  End;

Procedure TSocketStream.Skip (const Count : Integer);
  Begin
    Read (Count);
  End;

Function TSocketStream.Write (const Buf; Count : Integer) : LongInt;
const SendBlockSize = 2048;
type ByteBuffer = Array [1..2140000000] of Byte;
var I, L   : Integer;
    Buffer : ByteBuffer absolute Buf;
  Begin
    if Count <= 0 then
      begin
        Result := 0;
        exit;
      end;
    I := 1;
    While (I <= Count) and FSocket.Connected do
      begin
        L := Min (Count - I + 1, SendBlockSize);
        L := FSockStream.Write (Buffer [I], L);
        if L = 0 then
          break;
        Inc (I, L);
      end;
    Result := I - 1;
  End;

Procedure TSocketStream.Write (const Data : String);
var I, L : Integer;
  Begin
    L := Length (Data);
    I := Write (Data [1], L);
    if I < L then
      raise EWriteError.Create ('Could not write to socket');
  End;

Procedure TSocketStream.Write (const Data : TOctet);
  Begin
    FSockStream.WriteBuffer (Data, 1);
  End;

end.

